# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Repeater management view — shown in the content panel for repeater contacts."""

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
import contextlib
import copy

from gi.repository import Adw, Gio, GLib, Gtk
from meshy.models import Contact
from meshy.protocol import (
    build_send_cli_command,
)


def _fmt_uptime(secs):
    d, secs = divmod(secs, 86400)
    h, secs = divmod(secs, 3600)
    m, secs = divmod(secs, 60)
    parts = []
    if d:
        parts.append(_("{d}d").format(d=d))
    if h:
        parts.append(_("{h}h").format(h=h))
    if m:
        parts.append(_("{m}m").format(m=m))
    parts.append(_("{secs}s").format(secs=secs))
    return " ".join(parts)


def _fmt_ago(secs):
    if secs < 60:
        return _("{}s ago").format(secs)
    if secs < 3600:
        return _("{}m ago").format(secs // 60)
    if secs < 86400:
        return _("{}h ago").format(secs // 3600)
    return _("{}d ago").format(secs // 86400)


def _battery_pct(mv):
    if mv >= 4100:
        return 100
    if mv <= 3300:
        return 0
    return int((mv - 3300) / 8)


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/repeater-view.ui")
class RepeaterView(Gtk.Box):
    __gtype_name__ = "MeshyRepeaterView"

    REQUEST_TIMEOUT_MS = 15000
    FETCH_CMD_TIMEOUT_MS = 8000

    # Main
    _stack = Gtk.Template.Child("stack")

    # Status page
    _st_clock = Gtk.Template.Child("st_clock")
    _clock_drift_label = Gtk.Template.Child("clock_drift_label")
    _clock_sync_btn = Gtk.Template.Child("clock_sync_btn")
    _clock_reset_btn = Gtk.Template.Child("clock_reset_btn")
    _st_battery = Gtk.Template.Child("st_battery")
    _st_uptime = Gtk.Template.Child("st_uptime")
    _st_queue = Gtk.Template.Child("st_queue")
    _st_errors = Gtk.Template.Child("st_errors")
    _st_rssi = Gtk.Template.Child("st_rssi")
    _st_snr = Gtk.Template.Child("st_snr")
    _st_noise = Gtk.Template.Child("st_noise")
    _st_tx_air = Gtk.Template.Child("st_tx_air")
    _st_rx_air = Gtk.Template.Child("st_rx_air")
    _st_channel_util = Gtk.Template.Child("st_channel_util")
    _st_duty_cycle = Gtk.Template.Child("st_duty_cycle")
    _st_sent = Gtk.Template.Child("st_sent")
    _st_recv = Gtk.Template.Child("st_recv")
    _st_recv_errors = Gtk.Template.Child("st_recv_errors")
    _st_flood = Gtk.Template.Child("st_flood")
    _st_direct = Gtk.Template.Child("st_direct")
    _st_dups = Gtk.Template.Child("st_dups")
    _status_refresh_btn = Gtk.Template.Child("status_refresh_btn")

    # CLI page
    _cli_quick_box = Gtk.Template.Child("cli_quick_box")
    _cli_view = Gtk.Template.Child("cli_view")
    _cli_entry = Gtk.Template.Child("cli_entry")
    _cli_send_btn = Gtk.Template.Child("cli_send_btn")

    # Neighbors page
    _nb_spinner = Gtk.Template.Child("nb_spinner")
    _nb_status = Gtk.Template.Child("nb_status")
    _nb_list = Gtk.Template.Child("nb_list")
    _nb_refresh_btn = Gtk.Template.Child("nb_refresh_btn")
    _nb_load_more = Gtk.Template.Child("nb_load_more")
    _nb_map_btn = Gtk.Template.Child("nb_map_btn")

    # Access page
    _acl_spinner = Gtk.Template.Child("acl_spinner")
    _acl_status = Gtk.Template.Child("acl_status")
    _acl_list = Gtk.Template.Child("acl_list")
    _acl_refresh_btn = Gtk.Template.Child("acl_refresh_btn")
    _acl_add_btn = Gtk.Template.Child("acl_add_btn")

    # Regions page
    _reg_loading_box = Gtk.Template.Child("reg_loading_box")
    _reg_spinner = Gtk.Template.Child("reg_spinner")
    _reg_status = Gtk.Template.Child("reg_status")
    _reg_retry_btn = Gtk.Template.Child("reg_retry_btn")
    _reg_content_box = Gtk.Template.Child("reg_content_box")
    _reg_default_entry = Gtk.Template.Child("reg_default_entry")
    _reg_list = Gtk.Template.Child("reg_list")
    _reg_add_btn = Gtk.Template.Child("reg_add_btn")
    _reg_apply_btn = Gtk.Template.Child("reg_apply_btn")

    # Settings page – fetch spinners
    _basic_fetch_spinner = Gtk.Template.Child("basic_fetch_spinner")
    _loc_fetch_spinner = Gtk.Template.Child("loc_fetch_spinner")
    _radio_fetch_spinner = Gtk.Template.Child("radio_fetch_spinner")
    _advert_fetch_spinner = Gtk.Template.Child("advert_fetch_spinner")

    # Settings page
    _set_name = Gtk.Template.Child("set_name")
    _set_repeat = Gtk.Template.Child("set_repeat")
    _set_admin_pwd = Gtk.Template.Child("set_admin_pwd")
    _admin_pwd_save_btn = Gtk.Template.Child("admin_pwd_save_btn")
    _set_guest_pwd = Gtk.Template.Child("set_guest_pwd")
    _guest_pwd_save_btn = Gtk.Template.Child("guest_pwd_save_btn")
    _set_lat = Gtk.Template.Child("set_lat")
    _set_lon = Gtk.Template.Child("set_lon")
    _set_freq = Gtk.Template.Child("set_freq")
    _set_bw = Gtk.Template.Child("set_bw")
    _set_sf = Gtk.Template.Child("set_sf")
    _set_cr = Gtk.Template.Child("set_cr")
    _set_tx = Gtk.Template.Child("set_tx")
    _set_advert_int = Gtk.Template.Child("set_advert_int")
    _set_flood_int = Gtk.Template.Child("set_flood_int")
    _reboot_btn = Gtk.Template.Child("reboot_btn")
    _basic_fetch_btn = Gtk.Template.Child("basic_fetch_btn")
    _basic_save_btn = Gtk.Template.Child("basic_save_btn")
    _loc_fetch_btn = Gtk.Template.Child("loc_fetch_btn")
    _loc_save_btn = Gtk.Template.Child("loc_save_btn")
    _radio_fetch_btn = Gtk.Template.Child("radio_fetch_btn")
    _radio_save_btn = Gtk.Template.Child("radio_save_btn")
    _advert_fetch_btn = Gtk.Template.Child("advert_fetch_btn")
    _advert_save_btn = Gtk.Template.Child("advert_save_btn")

    # Telemetry
    _tele_btn = Gtk.Template.Child("tele_btn")

    # Bottom bar
    _logout_btn = Gtk.Template.Child("logout_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._contact = None
        self._timeout_id = None
        self._nb_offset = 0
        self._nb_total = 0
        self._nb_neighbors = []  # list of (contact_or_None, snr)
        self._acl_entries = []  # list of {'key': hex, 'perm': int}
        self._region_entries = []  # list of {'name', 'indent', 'home', 'flood'}
        self._original_region_entries = []
        self._original_default_scope = ""
        self._pending_region_cmds = []
        self._reg_timeout_id = None
        self._reg_cmd_ids = set()
        self._regions_pending = False
        self._reg_prefixes = []
        self._reg_popover_entry = None
        self._current_page = "placeholder"
        self._leaving_regions = False
        self._pending_cmd_ids = []
        self._pending_prefixes = []
        self._fetch_timeout_id = None
        self._fetch_btn = None
        self._fetch_spinner = None

        # CLI buffer reference
        self._cli_buf = self._cli_view.get_buffer()

        # Populate CLI quick command buttons
        for label, cmd in [
            ("get name", "get name"),
            ("get radio", "get radio"),
            ("get tx", "get tx"),
            ("ver", "ver"),
            ("neighbors", "neighbors"),
            ("advert", "advert"),
            ("clock", "clock"),
        ]:
            btn = Gtk.Button(label=label)
            btn.add_css_class("flat")
            btn.connect("clicked", lambda b, c=cmd: self._send_cli(c))
            self._cli_quick_box.append(btn)

        # Signal connections
        self._status_refresh_btn.connect("clicked", lambda *_: self._request_status())
        self._tele_btn.connect("clicked", lambda *_: self._request_telemetry())
        self._clock_sync_btn.connect("clicked", self._on_sync_time_clicked)
        self._clock_reset_btn.connect("clicked", self._on_clock_reset_clicked)

        self._cli_entry.connect(
            "activate", lambda *_: self._send_cli(self._cli_entry.get_text().strip())
        )
        self._cli_send_btn.connect(
            "clicked", lambda *_: self._send_cli(self._cli_entry.get_text().strip())
        )

        self._nb_refresh_btn.connect("clicked", lambda *_: self._request_neighbors(reset=True))
        self._nb_load_more.connect("clicked", lambda *_: self._request_neighbors(reset=False))
        self._nb_map_btn.connect("clicked", lambda *_: self._show_neighbors_map())

        self._acl_refresh_btn.connect("clicked", lambda *_: self._request_acl())
        self._acl_add_btn.connect("clicked", lambda *_: self._show_acl_add_dialog())

        self._reg_retry_btn.connect("clicked", lambda *_: self._request_regions())
        self._reg_add_btn.connect("clicked", lambda *_: self._show_region_add_dialog())
        self._reg_apply_btn.connect("clicked", lambda *_: self._on_reg_apply())
        self._reg_default_entry.connect("notify::text", lambda *_: self._on_reg_default_changed())
        self._stack.connect("notify::visible-child-name", self._on_stack_page_changed)

        reg_action_group = Gio.SimpleActionGroup()
        for name, cb in [
            ("toggle_flood", lambda *_: self._on_reg_toggle_flood()),
            ("set_home", lambda *_: self._on_reg_set_home()),
            ("remove", lambda *_: self._on_reg_remove()),
        ]:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", cb)
            reg_action_group.add_action(action)
        self._reg_list.insert_action_group("reg", reg_action_group)

        self._basic_fetch_btn.connect(
            "clicked",
            lambda *_: self._fetch_section(
                ["get name", "get repeat"],
                self._parse_basic,
                self._basic_fetch_btn,
                self._basic_fetch_spinner,
            ),
        )
        self._basic_save_btn.connect("clicked", lambda *_: self._save_basic())
        self._admin_pwd_save_btn.connect("clicked", lambda *_: self._save_admin_password())
        self._guest_pwd_save_btn.connect("clicked", lambda *_: self._save_guest_password())
        self._loc_fetch_btn.connect(
            "clicked",
            lambda *_: self._fetch_section(
                ["get lat", "get lon"],
                self._parse_location,
                self._loc_fetch_btn,
                self._loc_fetch_spinner,
            ),
        )
        self._loc_save_btn.connect("clicked", lambda *_: self._save_location())
        self._radio_fetch_btn.connect(
            "clicked",
            lambda *_: self._fetch_section(
                ["get radio", "get tx"],
                self._parse_radio,
                self._radio_fetch_btn,
                self._radio_fetch_spinner,
            ),
        )
        self._radio_save_btn.connect("clicked", lambda *_: self._save_radio())
        self._advert_fetch_btn.connect(
            "clicked",
            lambda *_: self._fetch_section(
                ["get advert.interval", "get flood.advert.interval"],
                self._parse_advert,
                self._advert_fetch_btn,
                self._advert_fetch_spinner,
            ),
        )
        self._advert_save_btn.connect("clicked", lambda *_: self._save_advert())
        self._reboot_btn.connect("activated", self._on_reboot)
        self._logout_btn.connect("clicked", self._on_logout)

    def _switch_tab(self, name):
        """Programmatically switch to a tab."""
        self._current_page = name
        self._stack.set_visible_child_name(name)

    def _start_timeout(self, on_timeout):
        """Start a request timeout. Cancels any previous one."""
        self._cancel_timeout()
        self._timeout_id = GLib.timeout_add(
            self.REQUEST_TIMEOUT_MS, lambda: (on_timeout(), False)[1]
        )

    def _cancel_timeout(self):
        if self._timeout_id:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

    # ─── Status ────────────────────────────────────────────────

    def _on_sync_time_clicked(self, btn):
        """Send the current system time to the repeater via CLI command."""
        if not self._contact:
            return
        import time
        from datetime import datetime

        ts = int(time.time())
        self._window.send_cli_command(self._contact, f"time {ts}")
        self._st_clock.set_subtitle(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        self._clock_drift_label.set_visible(False)
        self._clock_sync_btn.set_visible(False)
        self._clock_reset_btn.set_visible(False)
        self._window.show_toast(_("Time sync sent to repeater"))

    def _on_clock_reset_clicked(self, btn):
        """Reset the repeater's clock and reboot it (for when clock is ahead)."""
        if not self._contact:
            return
        dialog = Adw.AlertDialog(
            heading=_("Reset Clock & Reboot?"),
            body=_(
                "The repeater clock is ahead of the current time.\n"
                "Firmware does not allow setting time backwards.\n\n"
                "This will reset the clock and reboot the repeater."
            ),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("reset", _("Reset & Reboot"))
        dialog.set_response_appearance("reset", Adw.ResponseAppearance.DESTRUCTIVE)

        def _on_response(d, response):
            if response == "reset":
                self._window.send_cli_command(self._contact, "clkreboot")
                self._clock_drift_label.set_visible(False)
                self._clock_sync_btn.set_visible(False)
                self._clock_reset_btn.set_visible(False)
                self._st_clock.set_subtitle(_("Rebooting..."))
                self._window.show_toast(_("Clock reset & reboot sent to repeater"))

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _request_status(self):
        if not self._contact:
            return
        self._start_timeout(self._on_status_timeout)
        self._window.send_status_req(self._contact.public_key, self._on_status_response)

    def _on_status_timeout(self):
        self._st_battery.set_subtitle(_("Request timed out"))
        self._window.set_status_callback(None)
        self._start_pending_regions()

    def _on_status_response(self, status):
        self._cancel_timeout()
        mv = status["battery_mv"]
        pct = _battery_pct(mv)
        self._st_battery.set_subtitle(f"{pct}% ({mv} mV)")
        self._st_uptime.set_subtitle(_fmt_uptime(status["uptime_secs"]))
        self._st_queue.set_subtitle(str(status["queue_len"]))
        self._st_errors.set_subtitle(str(status["error_events"]))
        self._st_rssi.set_subtitle(f'{status["last_rssi"]} dB')
        self._st_snr.set_subtitle(f'{status["last_snr"]:.1f} dB')
        self._st_noise.set_subtitle(f'{status["noise_floor"]} dB')
        self._st_tx_air.set_subtitle(_fmt_uptime(status["tx_air_secs"]))
        self._st_rx_air.set_subtitle(_fmt_uptime(status["rx_air_secs"]))
        uptime = status["uptime_secs"]
        if uptime > 0:
            util = (status["tx_air_secs"] + status["rx_air_secs"]) / uptime * 100
            self._st_channel_util.set_subtitle(f"{util:.1f}%")
            duty = status["tx_air_secs"] / uptime * 100
            self._st_duty_cycle.set_subtitle(f"{duty:.1f}%")
        self._st_sent.set_subtitle(str(status["packets_sent"]))
        self._st_recv.set_subtitle(str(status["packets_recv"]))
        recv_errors = status.get("recv_errors")
        if recv_errors is not None:
            recv_errors_percentage = recv_errors / (status["packets_recv"] + recv_errors) * 100
            self._st_recv_errors.set_visible(True)
            self._st_recv_errors.set_subtitle(f"{recv_errors} / {recv_errors_percentage:.2f}%")
        self._st_flood.set_subtitle(f'TX: {status["flood_tx"]} / RX: {status["flood_rx"]}')
        self._st_direct.set_subtitle(f'TX: {status["direct_tx"]} / RX: {status["direct_rx"]}')
        self._st_dups.set_subtitle(f'Flood: {status["dup_flood"]} / Direct: {status["dup_direct"]}')
        self._start_pending_regions()

    def _start_pending_regions(self):
        if self._regions_pending:
            self._regions_pending = False
            GLib.timeout_add(500, lambda: (self._request_regions(), False)[1])

    # ─── Telemetry ─────────────────────────────────────────────

    def _request_telemetry(self):
        if not self._contact:
            return
        from meshy.models import ContactType

        self._tele_btn.set_sensitive(False)
        self._tele_btn.set_label(_("Requesting…"))

        timeout_id = [None]

        def _reset_btn():
            self._tele_btn.set_sensitive(True)
            self._tele_btn.set_label(_("Request Telemetry"))

        def on_timeout():
            _reset_btn()
            timeout_id[0] = None
            return False

        def on_response(lpp_data):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            _reset_btn()
            if not lpp_data:
                return
            from meshy.protocol import parse_cayenne_lpp

            entries = parse_cayenne_lpp(lpp_data)
            if entries:
                self._show_telemetry_dialog(entries)

        timeout_id[0] = GLib.timeout_add(self.REQUEST_TIMEOUT_MS, on_timeout)
        self._window.get_telemetry(
            self._contact.public_key, on_response, contact_type=ContactType.REPEATER
        )

    def _show_telemetry_dialog(self, entries: list[dict]):
        from meshy.models import battery_percent_from_mv
        from meshy.protocol import LPP_GPS, LPP_VOLTAGE

        gps_entry = next((e for e in entries if e["type"] == LPP_GPS), None)

        dialog = Adw.Dialog()
        dialog.set_title(_("Telemetry — {}").format(self._contact.name if self._contact else ""))
        dialog.set_content_width(420)
        dialog.set_content_height(550 if gps_entry else 400)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=12,
            margin_end=12,
            margin_top=12,
            margin_bottom=12,
        )

        channels = {}
        for entry in entries:
            channels.setdefault(entry["channel"], []).append(entry)

        for ch in sorted(channels):
            group = Adw.PreferencesGroup(title=_("Channel {}").format(ch))

            for entry in channels[ch]:
                val = entry["value"]
                if isinstance(val, dict):
                    alt_str = f', alt {val["alt"]:.0f} m' if abs(val["alt"]) > 0.01 else ""
                    subtitle = f'{val["lat"]:.4f}, {val["lon"]:.4f}{alt_str}'
                elif entry["type"] == LPP_VOLTAGE and ch == 1:
                    pct = battery_percent_from_mv(int(val * 1000))
                    subtitle = f"{pct}% ({val:.2f} V)"
                elif isinstance(val, float):
                    subtitle = f'{val:.2f} {entry["unit"]}'
                else:
                    subtitle = f'{val} {entry["unit"]}'

                title = _("Battery") if entry["type"] == LPP_VOLTAGE and ch == 1 else entry["name"]
                row = Adw.ActionRow(title=title, subtitle=subtitle)
                group.add(row)

            content.append(group)

        if gps_entry:
            from meshy.views import create_shumate_map, deregister_map_widget

            lat = gps_entry["value"]["lat"]
            lon = gps_entry["value"]["lon"]

            map_widget, marker_layer, viewport = create_shumate_map(zoom_level=14, lat=lat, lon=lon)
            if map_widget:
                map_widget.set_vexpand(False)
                map_widget.set_size_request(-1, 200)
                content.append(map_widget)
                dialog.connect("closed", lambda _d: deregister_map_widget(map_widget))

        scroll.set_child(content)
        toolbar_view.set_content(scroll)
        dialog.set_child(toolbar_view)
        dialog.present(self._window)

    # ─── CLI ───────────────────────────────────────────────────

    def _send_cli(self, cmd):
        if not cmd or not self._contact:
            return
        self._cli_entry.set_text("")
        self._cli_append(f"> {cmd}")
        self._window.send_frame(build_send_cli_command(self._contact.public_key, cmd))

    _CLI_MAX_LINES = 5000

    def _cli_append(self, text):
        end = self._cli_buf.get_end_iter()
        self._cli_buf.insert(end, text + "\n")
        line_count = self._cli_buf.get_line_count()
        if line_count > self._CLI_MAX_LINES:
            start = self._cli_buf.get_start_iter()
            trim_end = self._cli_buf.get_iter_at_line(line_count - self._CLI_MAX_LINES)
            self._cli_buf.delete(start, trim_end)
        GLib.idle_add(
            lambda: self._cli_view.get_parent()
            .get_vadjustment()
            .set_value(self._cli_view.get_parent().get_vadjustment().get_upper())
            or False
        )

    def on_cli_response(self, contact, message):
        """Called by window when a CLI response arrives."""
        if (
            self._contact
            and message.is_cli
            and contact.public_key_hex == self._contact.public_key_hex
        ):
            self._cli_append(message.text)

    # ─── Neighbors ─────────────────────────────────────────────

    def _request_neighbors(self, reset=True):
        if not self._contact:
            return
        if reset:
            self._nb_offset = 0
            self._nb_neighbors = []
            self._nb_map_btn.set_visible(False)
            child = self._nb_list.get_first_child()
            while child:
                next_c = child.get_next_sibling()
                self._nb_list.remove(child)
                child = next_c
        self._nb_spinner.set_spinning(True)
        self._nb_status.set_label(_("Requesting..."))
        self._nb_load_more.set_visible(False)
        self._start_timeout(self._on_neighbors_timeout)
        self._window.get_neighbors(
            self._contact.public_key, self._on_neighbors_response, offset=self._nb_offset
        )

    def _on_neighbors_timeout(self):
        self._nb_spinner.set_spinning(False)
        self._nb_status.set_label(_("Request timed out"))
        self._nb_load_more.set_visible(self._nb_offset < self._nb_total)
        self._window.set_neighbors_callback(None)

    def _on_neighbors_response(self, result):
        self._cancel_timeout()
        self._nb_spinner.set_spinning(False)
        self._nb_total = result["total"]
        neighbors = result["neighbors"]
        self._nb_offset += len(neighbors)

        for nb in neighbors:
            prefix = nb.get("pub_key_hex", "")
            last_heard = nb.get("last_heard", 0)
            snr = nb.get("snr", 0)

            contact = self._window.find_contact_by_key_prefix(prefix)
            self._nb_neighbors.append((contact, snr))
            title = contact.name if contact else prefix
            subtitle = f"{_fmt_ago(last_heard)} \u00b7 SNR: {snr:.1f} dB"
            row = Adw.ActionRow(title=title, subtitle=subtitle)
            row.add_prefix(Gtk.Image.new_from_icon_name("network-server-symbolic"))
            self._nb_list.append(row)

        self._nb_status.set_label(_("{} of {} neighbors").format(self._nb_offset, self._nb_total))
        self._nb_load_more.set_visible(self._nb_offset < self._nb_total)

        center = self._contact
        has_center = center and center.has_location
        has_located_nb = any(c and c.has_location for c, _ in self._nb_neighbors)
        self._nb_map_btn.set_visible(has_center and has_located_nb)

    def _show_neighbors_map(self):
        from gi.repository import Shumate
        from meshy.views.map_view import MapView

        center = self._contact
        if not center or not center.has_location:
            return

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=_("Neighbor Map")
        )

        # Remove marker_layer so we can re-add it on top of path layers
        map_widget.get_map().remove_layer(marker_layer)

        lats = [center.latitude]
        lons = [center.longitude]
        neighbors_located = []

        # First pass: add path layers (lines) so they render below markers
        for contact, snr in self._nb_neighbors:
            if not contact or not contact.has_location:
                continue
            neighbors_located.append((contact, snr))

            path_layer = MapView.create_dashed_path_layer(viewport)
            map_widget.get_map().add_layer(path_layer)

            center_node = Shumate.Marker()
            center_node.set_location(center.latitude, center.longitude)
            path_layer.add_node(center_node)

            nb_node = Shumate.Marker()
            nb_node.set_location(contact.latitude, contact.longitude)
            path_layer.add_node(nb_node)

            lats.append(contact.latitude)
            lons.append(contact.longitude)

        # Re-add marker layer on top of all path layers
        map_widget.get_map().add_layer(marker_layer)

        # Center repeater marker (red)
        center_widget = MapView._build_marker_widget(
            0.90, 0.16, 0.22, "network-server-symbolic", ""
        )
        center_widget.set_tooltip_text(center.name)
        center_marker = Shumate.Marker()
        center_marker.set_child(center_widget)
        center_marker.set_location(center.latitude, center.longitude)
        marker_layer.add_marker(center_marker)
        click = Gtk.GestureClick()
        click.connect(
            "pressed",
            lambda g, n, x, y: MapView.show_marker_popover(
                center_widget, center.name, center.latitude, center.longitude
            ),
        )
        center_widget.add_controller(click)

        # Neighbor markers (green) and SNR labels
        for contact, snr in neighbors_located:
            nb_widget = MapView._build_marker_widget(
                0.30, 0.69, 0.31, "network-server-symbolic", ""
            )
            nb_widget.set_tooltip_text(contact.name)
            nb_click = Gtk.GestureClick()
            nb_click.connect(
                "pressed",
                lambda g, n, x, y, w=nb_widget, c=contact, s=snr: MapView.show_marker_popover(
                    w, c.name, c.latitude, c.longitude, s
                ),
            )
            nb_widget.add_controller(nb_click)
            nb_marker = Shumate.Marker()
            nb_marker.set_child(nb_widget)
            nb_marker.set_location(contact.latitude, contact.longitude)
            marker_layer.add_marker(nb_marker)

            mid_lat = (center.latitude + contact.latitude) / 2
            mid_lon = (center.longitude + contact.longitude) / 2
            snr_widget = MapView.build_snr_widget(snr)
            snr_marker = Shumate.Marker()
            snr_marker.set_child(snr_widget)
            snr_marker.set_location(mid_lat, mid_lon)
            marker_layer.add_marker(snr_marker)

        MapView.fit_viewport(viewport, lats, lons)
        dialog.present(self._window)

    # ─── Access Control (ACL) ─────────────────────────────────

    def _parse_acl_response(self, data):
        """Parse raw ACL binary response into list of {'key': hex, 'perm': int}."""
        # Response: [code(1)][reserved(1)][tag(4)][payload...]
        # Payload: 7-byte entries: [pubkey_prefix(6)][perm(1)]
        buf = data[6:]  # skip code + reserved + tag
        entries = []
        i = 0
        while i + 7 <= len(buf):
            key = buf[i : i + 6].hex()
            perm = buf[i + 6]
            if key != "000000000000":
                entries.append({"key": key, "perm": perm})
            i += 7
        return entries

    def _request_acl(self):
        if not self._contact:
            return
        # Clear list
        child = self._acl_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._acl_list.remove(child)
            child = next_c
        self._acl_spinner.set_spinning(True)
        self._acl_status.set_label(_("Requesting..."))
        self._start_timeout(self._on_acl_timeout)
        self._window.get_acl(self._contact.public_key, self._on_acl_response)

    def _on_acl_timeout(self):
        self._acl_spinner.set_spinning(False)
        self._acl_status.set_label(_("Request timed out"))
        self._window.set_acl_callback(None)

    def _on_acl_response(self, data):
        self._cancel_timeout()
        self._acl_spinner.set_spinning(False)
        self._acl_entries = self._parse_acl_response(data)

        for entry in self._acl_entries:
            prefix = entry["key"]
            perm = entry["perm"]

            # Try to match to a known contact or own companion
            self_key = self._window.self_info.get("public_key", "")
            if self_key.startswith(prefix):
                name = self._window.self_info.get("name", "") or _("Me")
            else:
                contact = self._window.find_contact_by_key_prefix(prefix)
                name = contact.name if contact else None

            role = _("Admin") if (perm & 1) == 1 else _("Guest")
            title = name or prefix
            subtitle = f"{role} \u00b7 {prefix}"
            row = Adw.ActionRow(title=title, subtitle=subtitle)
            row.add_prefix(Gtk.Image.new_from_icon_name("avatar-default-symbolic"))

            # Remove button
            remove_btn = Gtk.Button(icon_name="user-trash-symbolic", valign=Gtk.Align.CENTER)
            remove_btn.add_css_class("destructive-action")
            remove_btn.connect("clicked", lambda b, k=prefix: self._on_acl_remove(k))
            row.add_suffix(remove_btn)
            self._acl_list.append(row)

        count = len(self._acl_entries)
        self._acl_status.set_label(
            _("{} entry").format(count) if count == 1 else _("{} entries").format(count)
        )

    def _on_acl_remove(self, key_hex):
        """Remove a node from ACL via setperm <key> -1."""
        if not self._contact:
            return
        # Find contact name for the confirmation message
        name = key_hex
        self_key = self._window.self_info.get("public_key", "")
        if self_key.startswith(key_hex):
            name = self._window.self_info.get("name", "") or _("Me")
        else:
            for c in self._window.contacts:
                if c.public_key_hex.startswith(key_hex):
                    name = c.name
                    break

        dialog = Adw.AlertDialog(
            heading=_("Remove from ACL?"),
            body=_("Remove {} from the access control list?").format(name),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("remove", _("Remove"))
        dialog.set_response_appearance("remove", Adw.ResponseAppearance.DESTRUCTIVE)

        def _on_response(d, response):
            if response == "remove":
                self._send_cmds([f"setperm {key_hex} -1"])
                # Refresh after a short delay to let the command take effect
                GLib.timeout_add(1000, lambda: (self._request_acl(), False)[1])

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _show_acl_add_dialog(self):
        """Show dialog to add a contact to the ACL."""
        if not self._contact:
            return

        dialog = Adw.AlertDialog(
            heading=_("Add to Access Control"),
            body=_("Select a contact and role."),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add"))
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        content.set_size_request(300, -1)

        # Role selector
        role_row = Adw.ComboRow(title=_("Role"))
        role_model = Gtk.StringList.new([_("Admin"), _("Guest")])
        role_row.set_model(role_model)
        role_row.set_selected(0)
        role_group = Adw.PreferencesGroup()
        role_group.add(role_row)
        content.append(role_group)

        # Contact list
        contacts_group = Adw.PreferencesGroup(title=_("Contact"))
        contact_scroll = Gtk.ScrolledWindow()
        contact_scroll.set_min_content_height(200)
        contact_scroll.set_max_content_height(300)
        contact_list = Gtk.ListBox()
        contact_list.set_selection_mode(Gtk.SelectionMode.SINGLE)
        contact_list.add_css_class("boxed-list")

        # Populate with contacts (exclude self repeater and non-chat contacts)
        acl_keys = {e["key"] for e in self._acl_entries}
        for c in self._window.contacts:
            # Skip the repeater itself
            if c.public_key_hex == self._contact.public_key_hex:
                continue
            # Skip contacts already in ACL
            if any(c.public_key_hex.startswith(k) for k in acl_keys):
                continue
            row = Adw.ActionRow(
                title=GLib.markup_escape_text(c.name), subtitle=c.public_key_hex[:12] + "..."
            )
            row._contact = c
            contact_list.append(row)

        list_wrapper = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            margin_start=8,
            margin_end=8,
            margin_top=8,
            margin_bottom=8,
        )
        list_wrapper.append(contact_list)
        contact_scroll.set_child(list_wrapper)
        contacts_group.add(contact_scroll)
        content.append(contacts_group)

        dialog.set_extra_child(content)

        def _on_response(d, response):
            if response != "add":
                return
            selected = contact_list.get_selected_row()
            if not selected or not hasattr(selected, "_contact"):
                return
            contact = selected._contact
            perm = 1 if role_row.get_selected() == 0 else 0
            self._send_cmds([f"setperm {contact.public_key_hex} {perm}"])
            # Refresh ACL after a delay
            GLib.timeout_add(1000, lambda: (self._request_acl(), False)[1])

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    # ─── Regions ───────────────────────────────────────────────

    def _cancel_reg_timeout(self):
        if self._reg_timeout_id:
            GLib.source_remove(self._reg_timeout_id)
            self._reg_timeout_id = None

    def _request_regions(self):
        if not self._contact:
            return
        # Cancel previous region fetch
        self._cancel_reg_timeout()
        for src_id in list(self._reg_cmd_ids):
            GLib.source_remove(src_id)
        self._window.cancel_cli_prefixes(self._reg_prefixes)
        self._reg_cmd_ids = set()
        self._reg_prefixes = []
        # Clear region list
        child = self._reg_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._reg_list.remove(child)
            child = next_c
        # Show loading, hide content
        self._reg_loading_box.set_visible(True)
        self._reg_content_box.set_visible(False)
        self._reg_spinner.set_spinning(True)
        self._reg_status.set_label(_("Loading regions..."))
        self._reg_retry_btn.set_visible(False)

        commands = ["region", "region default"]
        responses = [None, None]
        remaining = [2]

        def make_cb(idx):
            def on_response(text):
                responses[idx] = text.strip()
                remaining[0] -= 1
                if remaining[0] <= 0:
                    self._cancel_reg_timeout()
                    self._on_regions_loaded(responses)

            return on_response

        def on_timeout():
            self._reg_timeout_id = None
            self._window.cancel_cli_prefixes(self._reg_prefixes)
            self._reg_prefixes = []
            received = [r for r in responses if r is not None]
            if received:
                self._on_regions_loaded(responses)
            else:
                self._on_regions_timeout()

        self._reg_timeout_id = GLib.timeout_add(
            self.REQUEST_TIMEOUT_MS, lambda: (on_timeout(), False)[1]
        )

        for i, cmd in enumerate(commands):
            holder = [0]

            def send(c=cmd, idx=i, h=holder):
                self._reg_cmd_ids.discard(h[0])
                prefix = f"{self._window.cli_prefix_counter:02X}|"
                self._reg_prefixes.append(prefix)
                self._window.send_cli_with_prefix(self._contact.public_key, c, make_cb(idx))
                return False

            holder[0] = GLib.timeout_add(i * 1500, send)
            self._reg_cmd_ids.add(holder[0])

    def _on_regions_timeout(self):
        self._reg_spinner.set_spinning(False)
        self._reg_status.set_label(_("Request timed out"))
        self._reg_retry_btn.set_visible(True)

    def _on_regions_loaded(self, responses):
        self._region_entries = []

        region_text = responses[0] if len(responses) > 0 and responses[0] else ""
        default_text = responses[1] if len(responses) > 1 and responses[1] else ""

        if not region_text:
            self._on_regions_timeout()
            return

        # Parse default scope
        default_name = ""
        if "default scope is" in default_text:
            val = default_text.split("default scope is", 1)[1].strip()
            if val and val != "<null>":
                default_name = val
        self._reg_default_entry.set_text(default_name)
        self._original_default_scope = default_name

        # Parse region tree
        for line in region_text.split("\n"):
            if not line.strip():
                continue
            indent = len(line) - len(line.lstrip(" "))
            stripped = line.strip()
            flood = stripped.endswith(" F")
            if flood:
                stripped = stripped[:-2].strip()
            home = stripped.endswith("^")
            if home:
                stripped = stripped[:-1]
            self._region_entries.append(
                {
                    "name": stripped,
                    "indent": indent,
                    "home": home,
                    "flood": flood,
                }
            )

        self._original_region_entries = copy.deepcopy(self._region_entries)
        self._pending_region_cmds = []
        self._reg_apply_btn.set_sensitive(False)

        self._rebuild_region_rows()
        self._reg_loading_box.set_visible(False)
        self._reg_content_box.set_visible(True)

    def _rebuild_region_rows(self):
        child = self._reg_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._reg_list.remove(child)
            child = next_c

        for entry in self._region_entries:
            is_wildcard = entry["name"] == "*"
            title = _("Global (wildcard)") if is_wildcard else entry["name"]
            parts = []
            if entry["home"]:
                parts.append(_("Home"))
            parts.append(_("Flood allowed") if entry["flood"] else _("Flood denied"))
            subtitle = " · ".join(parts)

            row = Adw.ActionRow(
                title=GLib.markup_escape_text(title),
                subtitle=subtitle,
            )

            prefix_box = Gtk.Box(spacing=0)
            visual_indent = max(0, entry["indent"] - 1)
            if visual_indent > 0:
                spacer = Gtk.Box()
                spacer.set_size_request(visual_indent * 16, -1)
                prefix_box.append(spacer)
            flood_icon = "check-plain-symbolic" if entry["flood"] else "circle-crossed-symbolic"
            prefix_box.append(Gtk.Image.new_from_icon_name(flood_icon))
            row.add_prefix(prefix_box)

            menu_btn = Gtk.Button(
                icon_name="view-more-symbolic",
                valign=Gtk.Align.CENTER,
                tooltip_text=_("Actions"),
            )
            menu_btn.add_css_class("flat")
            menu_btn.connect("clicked", lambda b, e=entry: self._show_reg_popover(b, e))
            row.add_suffix(menu_btn)

            self._reg_list.append(row)

    def _show_reg_popover(self, button, entry):
        self._reg_popover_entry = entry
        is_wildcard = entry["name"] == "*"
        menu = Gio.Menu()
        flood_label = _("Deny Flood") if entry["flood"] else _("Allow Flood")
        menu.append(flood_label, "reg.toggle_flood")
        if not is_wildcard:
            if not entry["home"]:
                menu.append(_("Set as Home"), "reg.set_home")
            menu.append(_("Remove"), "reg.remove")
        popover = Gtk.PopoverMenu(menu_model=menu)
        popover.set_has_arrow(False)
        popover.set_parent(button)
        popover.connect("closed", lambda p: GLib.idle_add(p.unparent))
        popover.popup()

    def _mark_regions_dirty(self):
        self._reg_apply_btn.set_sensitive(True)

    def _check_regions_dirty(self):
        if self._pending_region_cmds:
            return True
        return self._reg_default_entry.get_text().strip() != self._original_default_scope

    def _on_reg_default_changed(self):
        if self._check_regions_dirty():
            self._mark_regions_dirty()
        else:
            self._reg_apply_btn.set_sensitive(False)

    def _on_reg_toggle_flood(self):
        entry = self._reg_popover_entry
        if not entry:
            return
        entry["flood"] = not entry["flood"]
        cmd = "region allowf" if entry["flood"] else "region denyf"
        self._pending_region_cmds.append(f'{cmd} {entry["name"]}')
        self._rebuild_region_rows()
        self._mark_regions_dirty()

    def _on_reg_set_home(self):
        entry = self._reg_popover_entry
        if not entry:
            return
        for e in self._region_entries:
            e["home"] = False
        entry["home"] = True
        self._pending_region_cmds.append(f'region home {entry["name"]}')
        self._rebuild_region_rows()
        self._mark_regions_dirty()

    def _on_reg_remove(self):
        entry = self._reg_popover_entry
        if not entry:
            return
        idx = self._region_entries.index(entry)
        if (
            idx + 1 < len(self._region_entries)
            and self._region_entries[idx + 1]["indent"] > entry["indent"]
        ):
            self._window.show_toast(_("Remove child regions first"))
            return
        self._region_entries.remove(entry)
        self._pending_region_cmds.append(f'region remove {entry["name"]}')
        self._rebuild_region_rows()
        self._mark_regions_dirty()

    def _on_reg_apply(self):
        cmds = list(self._pending_region_cmds)
        current_default = self._reg_default_entry.get_text().strip()
        if current_default != self._original_default_scope:
            cmd = (
                f"region default {current_default}" if current_default else "region default <null>"
            )
            cmds.append(cmd)
        if not cmds:
            return
        cmds.append("region save")
        self._send_cmds(cmds)
        self._pending_region_cmds = []
        self._reg_apply_btn.set_sensitive(False)
        GLib.timeout_add(1500, lambda: (self._request_regions(), False)[1])

    def _on_stack_page_changed(self, stack, pspec):
        new_page = stack.get_visible_child_name()
        if self._leaving_regions:
            self._leaving_regions = False
            self._current_page = new_page
            return
        prev_page = self._current_page
        if prev_page != "regions" or not self._check_regions_dirty():
            self._current_page = new_page
            return
        self._leaving_regions = True
        stack.set_visible_child_name("regions")
        target_page = new_page

        dialog = Adw.AlertDialog(
            heading=_("Unsaved Changes"),
            body=_("You have unsaved region changes."),
        )
        dialog.add_response("discard", _("Discard"))
        dialog.add_response("apply", _("Apply"))
        dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.set_response_appearance("apply", Adw.ResponseAppearance.SUGGESTED)
        dialog.set_close_response("close")

        def _on_response(d, response):
            if response == "apply":
                self._on_reg_apply()
                self._leaving_regions = True
                stack.set_visible_child_name(target_page)
            elif response == "discard":
                self._pending_region_cmds = []
                self._reg_apply_btn.set_sensitive(False)
                self._request_regions()
                self._leaving_regions = True
                stack.set_visible_child_name(target_page)
            else:
                self._leaving_regions = False

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _show_region_add_dialog(self):
        if not self._contact:
            return

        dialog = Adw.AlertDialog(
            heading=_("Add Region"),
            body=_("Enter a name and optionally choose a parent region."),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add"))
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        content.set_size_request(300, -1)

        name_entry = Adw.EntryRow(title=_("Region Name"))
        name_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        name_list.add_css_class("boxed-list")
        name_list.append(name_entry)
        content.append(name_list)

        parent_names = (
            [_("None")] + ["*"] + [e["name"] for e in self._region_entries if e["name"] != "*"]
        )
        parent_row = Adw.ComboRow(title=_("Parent"))
        parent_model = Gtk.StringList.new(parent_names)
        parent_row.set_model(parent_model)
        parent_row.set_selected(0)
        parent_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        parent_list.add_css_class("boxed-list")
        parent_list.append(parent_row)
        content.append(parent_list)

        dialog.set_extra_child(content)

        def _on_response(d, response):
            if response != "add":
                return
            name = name_entry.get_text().strip()
            if not name:
                return
            selected = parent_row.get_selected()
            if selected == 0:
                parent_entry = None
            else:
                parent_name = parent_names[selected]
                parent_entry = next(
                    (e for e in self._region_entries if e["name"] == parent_name), None
                )
            child_indent = (parent_entry["indent"] + 1) if parent_entry else 1
            insert_idx = len(self._region_entries)
            if parent_entry:
                pi = self._region_entries.index(parent_entry)
                insert_idx = pi + 1
                while (
                    insert_idx < len(self._region_entries)
                    and self._region_entries[insert_idx]["indent"] > parent_entry["indent"]
                ):
                    insert_idx += 1
            self._region_entries.insert(
                insert_idx,
                {
                    "name": name,
                    "indent": child_indent,
                    "home": False,
                    "flood": True,
                },
            )
            if parent_entry:
                self._pending_region_cmds.append(f"region put {name} {parent_name}")
            else:
                self._pending_region_cmds.append(f"region put {name}")
            self._rebuild_region_rows()
            self._mark_regions_dirty()

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    # ─── Settings ──────────────────────────────────────────────

    def _cancel_fetch(self):
        """Cancel any in-progress settings fetch."""
        if self._fetch_timeout_id:
            GLib.source_remove(self._fetch_timeout_id)
            self._fetch_timeout_id = None
        for src_id in self._pending_cmd_ids:
            GLib.source_remove(src_id)
        self._window.cancel_cli_prefixes(self._pending_prefixes)
        self._pending_cmd_ids = []
        self._pending_prefixes = []
        if self._fetch_spinner:
            self._fetch_spinner.set_spinning(False)
            self._fetch_spinner.set_visible(False)
        if self._fetch_btn:
            self._fetch_btn.set_sensitive(True)
        self._fetch_btn = None
        self._fetch_spinner = None

    def _fetch_section(self, commands, parser, fetch_btn, spinner):
        """Fetch a section's settings via sequential CLI commands with retry."""
        if not self._contact:
            return
        self._cancel_fetch()
        self._fetch_btn = fetch_btn
        self._fetch_spinner = spinner
        fetch_btn.set_sensitive(False)
        spinner.set_visible(True)
        spinner.set_spinning(True)
        responses = [None] * len(commands)
        self._fetch_next_cmd(commands, responses, 0, parser, False)

    def _fetch_next_cmd(self, commands, responses, idx, parser, retried):
        """Send command at idx; on response advance, on timeout retry once."""
        if not self._contact:
            self._cancel_fetch()
            return
        if idx >= len(commands):
            self._cancel_fetch()
            parser(responses)
            return

        def on_response(text):
            if self._fetch_timeout_id:
                GLib.source_remove(self._fetch_timeout_id)
                self._fetch_timeout_id = None
            responses[idx] = text.strip()
            self._pending_prefixes = []
            self._fetch_next_cmd(commands, responses, idx + 1, parser, False)

        def on_timeout():
            self._fetch_timeout_id = None
            self._window.cancel_cli_prefixes(self._pending_prefixes)
            self._pending_prefixes = []
            if not retried:
                self._fetch_next_cmd(commands, responses, idx, parser, True)
            else:
                self._fetch_next_cmd(commands, responses, idx + 1, parser, False)

        self._fetch_timeout_id = GLib.timeout_add(
            self.FETCH_CMD_TIMEOUT_MS, lambda: (on_timeout(), False)[1]
        )
        prefix = f"{self._window.cli_prefix_counter:02X}|"
        self._pending_prefixes = [prefix]
        self._window.send_cli_with_prefix(self._contact.public_key, commands[idx], on_response)

    def _parse_basic(self, responses):
        for text in responses:
            if text is None:
                continue
            t = text.strip().lstrip("> ").strip()
            low = t.lower()
            if low in ("on", "off", "1", "0", "true", "false"):
                self._set_repeat.set_active(low in ("on", "1", "true"))
            elif t and "," not in t and not any(c in t for c in "{}[]"):
                self._set_name.set_text(t)

    def _parse_location(self, responses):
        for i, text in enumerate(responses):
            if text is None:
                continue
            t = text.strip().lstrip("> ").strip()
            try:
                val = float(t)
                if i == 0:
                    self._set_lat.set_text(str(val))
                else:
                    self._set_lon.set_text(str(val))
            except ValueError:
                pass

    def _parse_radio(self, responses):
        for text in responses:
            if text is None:
                continue
            t = text.strip().lstrip("> ").strip()
            if "," in t:
                parts = t.split(",")
                try:
                    self._set_freq.set_text(parts[0].strip())
                    self._set_bw.set_text(parts[1].strip())
                    self._set_sf.set_value(int(float(parts[2].strip())))
                    if len(parts) > 3:
                        self._set_cr.set_value(int(float(parts[3].strip())))
                except (ValueError, IndexError):
                    pass
            else:
                with contextlib.suppress(ValueError):
                    self._set_tx.set_value(int(float(t)))

    def _parse_advert(self, responses):
        for i, text in enumerate(responses):
            if text is None:
                continue
            t = text.strip().lstrip("> ").strip()
            try:
                val = int(float(t))
                if i == 0:
                    self._set_advert_int.set_value(val)
                else:
                    self._set_flood_int.set_value(val)
            except ValueError:
                pass

    def _save_admin_password(self):
        if not self._contact:
            return
        pwd = self._set_admin_pwd.get_text()
        if not pwd:
            return
        self._send_cmds([f"password {pwd}"])
        self._set_admin_pwd.set_text("")
        self._window.show_toast(_("Admin password changed — re-login required"))

    def _save_guest_password(self):
        if not self._contact:
            return
        pwd = self._set_guest_pwd.get_text()
        self._send_cmds([f"set guest.password {pwd}"])

    def _save_basic(self):
        if not self._contact:
            return
        cmds = []
        name = self._set_name.get_text().strip()
        if name:
            cmds.append(f"set name {name}")
        cmds.append(f'set repeat {"on" if self._set_repeat.get_active() else "off"}')
        self._send_cmds(cmds)

    def _save_location(self):
        if not self._contact:
            return
        cmds = []
        with contextlib.suppress(ValueError):
            cmds.append(f"set lat {float(self._set_lat.get_text())}")
        with contextlib.suppress(ValueError):
            cmds.append(f"set lon {float(self._set_lon.get_text())}")
        self._send_cmds(cmds)

    def _save_radio(self):
        if not self._contact:
            return
        cmds = []
        freq = self._set_freq.get_text().strip()
        bw = self._set_bw.get_text().strip()
        if freq and bw:
            sf = int(self._set_sf.get_value())
            cr = int(self._set_cr.get_value())
            cmds.append(f"set radio {freq},{bw},{sf},{cr}")
        tx = int(self._set_tx.get_value())
        if tx > 0:
            cmds.append(f"set tx {tx}")
        self._send_cmds(cmds)

    def _save_advert(self):
        if not self._contact:
            return
        cmds = [
            f"set advert.interval {int(self._set_advert_int.get_value())}",
            f"set flood.advert.interval {int(self._set_flood_int.get_value())}",
        ]
        self._send_cmds(cmds)

    def _send_cmds(self, cmds):
        if not cmds:
            return
        for i, cmd in enumerate(cmds):
            GLib.timeout_add(
                i * 300,
                lambda c=cmd: (
                    self._window.send_frame(build_send_cli_command(self._contact.public_key, c))
                    or False
                ),
            )
        name = self._contact.name if self._contact else "repeater"
        self._window.show_toast(_("Settings sent to {}").format(name))

    def _on_reboot(self, *args):
        confirm = Adw.AlertDialog(
            heading=_("Reboot Repeater?"),
            body=_("Reboot {}? It will be offline briefly.").format(self._contact.name),
        )
        confirm.add_response("cancel", _("Cancel"))
        confirm.add_response("reboot", _("Reboot"))
        confirm.set_response_appearance("reboot", Adw.ResponseAppearance.DESTRUCTIVE)
        confirm.connect(
            "response", lambda d, r: (self._send_cli("reboot")) if r == "reboot" else None
        )
        confirm.present(self._window)

    # ─── Public API ────────────────────────────────────────────

    def _reset_ui(self):
        """Reset all UI widgets to default/empty state."""
        self._cancel_fetch()

        # Status rows
        for w in (
            self._st_battery,
            self._st_uptime,
            self._st_queue,
            self._st_errors,
            self._st_rssi,
            self._st_snr,
            self._st_noise,
            self._st_tx_air,
            self._st_rx_air,
            self._st_sent,
            self._st_recv,
            self._st_flood,
            self._st_direct,
            self._st_dups,
        ):
            w.set_subtitle("--")
        self._st_recv_errors.set_visible(False)

        # Neighbors
        child = self._nb_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._nb_list.remove(child)
            child = next_c
        self._nb_offset = 0
        self._nb_total = 0
        self._nb_neighbors = []
        self._nb_spinner.set_spinning(False)
        self._nb_status.set_label("")
        self._nb_load_more.set_visible(False)
        self._nb_map_btn.set_visible(False)

        # ACL
        child = self._acl_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._acl_list.remove(child)
            child = next_c
        self._acl_entries = []
        self._acl_spinner.set_spinning(False)
        self._acl_status.set_label("")

        # Regions
        child = self._reg_list.get_first_child()
        while child:
            next_c = child.get_next_sibling()
            self._reg_list.remove(child)
            child = next_c
        self._region_entries = []
        self._original_region_entries = []
        self._original_default_scope = ""
        self._reg_default_entry.set_text("")
        self._reg_loading_box.set_visible(True)
        self._reg_content_box.set_visible(False)
        self._reg_spinner.set_spinning(False)
        self._reg_status.set_label("")
        self._reg_retry_btn.set_visible(False)

        # Settings
        self._set_name.set_text("")
        self._set_repeat.set_active(False)
        self._set_admin_pwd.set_text("")
        self._set_guest_pwd.set_text("")
        self._set_lat.set_text("")
        self._set_lon.set_text("")
        self._set_freq.set_text("")
        self._set_bw.set_text("")
        self._set_sf.set_value(5)
        self._set_cr.set_value(5)
        self._set_tx.set_value(1)
        self._set_advert_int.set_value(0)
        self._set_flood_int.set_value(0)

        # CLI
        self._cli_buf.set_text("")

    def set_repeater(self, contact: Contact, login_timestamp=None, is_admin=True):
        """Set the active repeater and switch to status view."""
        # Clean up previous repeater state
        self._cancel_timeout()
        self._cancel_reg_timeout()
        self._pending_region_cmds = []
        self._leaving_regions = False
        self._pending_cmd_ids = []
        self._pending_prefixes = []
        self._window.set_cli_response_callback(None)
        self._window.set_status_callback(None)
        self._window.set_neighbors_callback(None)
        self._window.set_acl_callback(None)
        self._reset_ui()

        self._contact = contact
        self._is_admin = is_admin

        # Show/hide admin-only tabs
        for name in ("cli", "access", "settings", "regions"):
            child = self._stack.get_child_by_name(name)
            if child:
                self._stack.get_page(child).set_visible(is_admin)

        # Show/hide clock sync controls (admin only)
        self._st_clock.set_visible(is_admin)

        # Display the repeater's clock time from the login response
        if is_admin and login_timestamp:
            import time
            from datetime import datetime

            dt = datetime.fromtimestamp(login_timestamp)
            self._st_clock.set_subtitle(dt.strftime("%Y-%m-%d %H:%M:%S"))
            now = int(time.time())
            drift = now - login_timestamp  # positive = behind, negative = ahead
            if abs(drift) > 300:  # more than 5 minutes off
                self._clock_drift_label.set_label(_("Time is off"))
                self._clock_drift_label.set_visible(True)
                if drift > 0:
                    # Repeater clock is behind — can set time forward
                    self._clock_sync_btn.set_visible(True)
                    self._clock_reset_btn.set_visible(False)
                else:
                    # Repeater clock is ahead — firmware rejects setting time back
                    self._clock_sync_btn.set_visible(False)
                    self._clock_reset_btn.set_visible(True)
            else:
                self._clock_drift_label.set_visible(False)
                self._clock_sync_btn.set_visible(False)
                self._clock_reset_btn.set_visible(False)
        else:
            self._st_clock.set_subtitle("--")
            self._clock_drift_label.set_visible(False)
            self._clock_sync_btn.set_visible(False)
            self._clock_reset_btn.set_visible(False)
        self._switch_tab("status")
        # Clear CLI output
        self._cli_buf.set_text("")
        # Register CLI callback for admin
        if is_admin:
            self._window.set_cli_response_callback(self.on_cli_response)
            self._regions_pending = True
        # Request status (regions will load after status completes)
        self._request_status()

    def _on_logout(self, *args):
        if self._contact:
            self._window.logout_repeater(self._contact.public_key_hex)

    def clear(self):
        self._cancel_timeout()
        self._cancel_reg_timeout()
        self._cancel_fetch()
        self._pending_region_cmds = []
        self._reg_apply_btn.set_sensitive(False)
        self._contact = None
        self._current_page = "placeholder"
        self._stack.set_visible_child_name("placeholder")
        self._window.set_cli_response_callback(None)
        self._window.set_status_callback(None)
        self._window.set_neighbors_callback(None)
        self._window.set_acl_callback(None)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Gdk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, Gdk, Gtk
from meshy.theme_manager import THEMES, ThemeManager


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/theme-chooser-dialog.ui")
class ThemeChooserDialog(Adw.Dialog):
    __gtype_name__ = "MeshyThemeChooserDialog"

    flow_box = Gtk.Template.Child()

    def __init__(self, settings, on_theme_changed_cb, **kwargs):
        super().__init__(**kwargs)

        self._settings = settings
        self._on_theme_changed_cb = on_theme_changed_cb

        current = settings.get_string("custom-theme")
        self._buttons = {}
        self._tile_provider = Gtk.CssProvider()
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(),
            self._tile_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        css_parts = []
        for theme_id, theme in THEMES.items():
            btn = self._create_tile(theme_id, theme, theme_id == current)
            self.flow_box.append(btn)
            self._buttons[theme_id] = btn
            css_parts.append(self._tile_css(theme_id, theme))

        self._tile_provider.load_from_string("\n".join(css_parts))

    def _create_tile(self, theme_id, theme, selected):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        color_box = Gtk.Box(hexpand=True, vexpand=True)
        color_box.set_size_request(100, 50)
        color_box.add_css_class(f"theme-preview-{theme_id}")
        color_box.add_css_class("theme-tile")

        if theme.get("accent"):
            accent_bar = Gtk.Box(hexpand=True, height_request=6)
            accent_bar.add_css_class(f"theme-accent-{theme_id}")
            accent_bar.set_valign(Gtk.Align.END)
            overlay = Gtk.Overlay()
            overlay.set_child(color_box)
            overlay.add_overlay(accent_bar)
            box.append(overlay)
        else:
            box.append(color_box)

        label = Gtk.Label(label=theme["name"])
        label.add_css_class("caption")
        box.append(label)

        btn = Gtk.ToggleButton()
        btn.set_child(box)
        btn.add_css_class("flat")
        btn.set_active(selected)
        btn.connect("toggled", self._on_toggled, theme_id)

        return btn

    def _on_toggled(self, button, theme_id):
        if not button.get_active():
            return
        for tid, btn in self._buttons.items():
            if tid != theme_id:
                btn.set_active(False)
        ThemeManager.get_default().set_custom_theme(theme_id)
        if self._on_theme_changed_cb:
            self._on_theme_changed_cb()

    @staticmethod
    def _tile_css(theme_id, theme):
        if theme.get("background"):
            r, g, b = theme["background"][:3]
            bg = f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"
        else:
            bg = "@window_bg_color"

        if theme.get("foreground"):
            r, g, b = theme["foreground"][:3]
            fg = f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"
        else:
            fg = "@window_fg_color"

        parts = [f".theme-preview-{theme_id} {{ background-color: {bg}; color: {fg}; }}"]

        if theme.get("accent"):
            r, g, b = theme["accent"][:3]
            accent = f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"
            parts.append(
                f".theme-accent-{theme_id} {{ background-color: {accent}; border-radius: 0 0 12px 12px; }}"
            )

        return "\n".join(parts)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared helpers for chat_view and channels_view."""

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Shumate", "1.0")
from gi.repository import Adw, Gio, GObject, Gtk, Shumate
from meshy.views import (
    _CONTACT_TYPE_ICONS,
    _contact_pick_factory_bind,
    _contact_pick_factory_setup,
    _contact_pick_factory_unbind,
    _ContactPickItem,
    create_shumate_map,
)

CONTACT_TYPE_ICONS = _CONTACT_TYPE_ICONS


def ensure_loc_map(outer, lat, lon):
    """Create or update the inline location map widget on a chat item."""
    if outer._loc_map_widget is not None:
        outer._loc_viewport.set_location(lat, lon)
        outer._loc_marker.set_location(lat, lon)
        outer._loc_map_frame.set_child(outer._loc_map_widget)
        return

    from meshy.views.map_view import MapView

    loc_map, loc_marker_layer, loc_viewport = create_shumate_map(zoom_level=14)
    loc_map.set_vexpand(False)
    loc_map.set_hexpand(True)
    loc_map.set_size_request(100, 150)
    loc_map.get_scale().set_visible(False)
    loc_map.set_show_zoom_buttons(False)
    loc_map.get_license().set_visible(False)

    loc_marker_widget = MapView._build_marker_widget(
        0.26, 0.52, 0.96, "find-location-symbolic", None, show_label=False
    )
    loc_marker = Shumate.Marker()
    loc_marker.set_child(loc_marker_widget)
    loc_marker_layer.add_marker(loc_marker)

    loc_viewport.set_zoom_level(14)
    loc_viewport.set_location(lat, lon)
    loc_marker.set_location(lat, lon)
    outer._loc_map_widget = loc_map
    outer._loc_viewport = loc_viewport
    outer._loc_marker = loc_marker
    outer._loc_map_frame.set_child(loc_map)


def show_location_map(window, lat, lon):
    """Open a dialog with the shared location on a map."""
    from meshy.views.map_view import MapView

    dialog, map_w, m_layer, vp = MapView.open_map_dialog(window, _("Shared Location"))
    marker = MapView.create_marker(
        0.26, 0.52, 0.96, "find-location-symbolic", _("Shared"), lat, lon
    )
    m_layer.add_marker(marker)

    dev = window.device_info
    if dev and dev.latitude is not None and dev.longitude is not None:
        self_marker = MapView.create_marker(
            0.90,
            0.16,
            0.22,
            "avatar-default-symbolic",
            dev.name or _("My Device"),
            dev.latitude,
            dev.longitude,
        )
        m_layer.add_marker(self_marker)

    vp.set_zoom_level(14)
    vp.set_location(lat, lon)
    dialog.present(window)


def open_location_external(btn, lat, lon):
    """Launch external map app for the given coordinates."""
    launcher = Gtk.UriLauncher.new(f"geo:{lat},{lon}")
    launcher.launch(None, None, None, None)


def share_contact_dialog(window, contacts, send_callback):
    """Show a contact picker dialog and call send_callback(contact) on selection."""
    if not contacts:
        return

    builder = Gtk.Builder.new_from_resource(
        "/page/codeberg/sesivany/Meshy/ui/share-contact-dialog.ui"
    )
    dialog = builder.get_object("share_dialog")
    search_btn = builder.get_object("share_search_btn")
    search_entry = builder.get_object("share_search_entry")
    list_view = builder.get_object("share_list_view")
    status_label = builder.get_object("share_status_label")

    search_btn.bind_property("active", search_entry, "visible", GObject.BindingFlags.SYNC_CREATE)
    search_entry.connect("stop-search", lambda _e: search_btn.set_active(False))

    store = Gio.ListStore(item_type=_ContactPickItem)
    for c in sorted(contacts, key=lambda c: c.name.lower()):
        store.append(_ContactPickItem(c))

    custom_filter = Gtk.CustomFilter.new(_share_filter_func, search_entry)

    def _on_search_changed(*_args):
        custom_filter.changed(Gtk.FilterChange.DIFFERENT)

    search_entry.connect("search-changed", _on_search_changed)

    filter_model = Gtk.FilterListModel(model=store, filter=custom_filter)
    selection = Gtk.NoSelection(model=filter_model)
    list_view.set_model(selection)

    def _on_share(contact):
        send_callback(contact)
        dialog.close()

    factory = Gtk.SignalListItemFactory()
    factory.connect("setup", lambda f, li: _contact_pick_factory_setup(f, li, _("Share")))
    factory.connect("bind", lambda f, li: _contact_pick_factory_bind(f, li, _on_share))
    factory.connect("unbind", _contact_pick_factory_unbind)
    list_view.set_factory(factory)

    def _on_activate(lv, pos):
        item = filter_model.get_item(pos)
        if item:
            _on_share(item.contact)

    list_view.connect("activate", _on_activate)

    total = store.get_n_items()
    status_label.set_label(_("{total} contacts").format(total=total) if total else _("No contacts"))

    dialog._refs = (store, custom_filter, filter_model, selection, factory)
    dialog.present(window)


def _share_filter_func(item, search_entry):
    q = search_entry.get_text().lower()
    return not q or q in item.contact.name.lower()


def share_location(window, send_callback):
    """Share the device's current GPS coordinates via send_callback(text)."""
    dev = window.device_info
    if not dev or dev.latitude is None or dev.longitude is None:
        dialog = Adw.AlertDialog(
            heading=_("Location Not Available"),
            body=_("Your companion does not have a GPS location."),
        )
        dialog.add_response("ok", _("OK"))
        dialog.present(window)
        return
    send_callback(f"{dev.latitude:.6f},{dev.longitude:.6f}")


def share_location_map(window, send_callback):
    """Open a map picker to choose and share a location via send_callback(text)."""
    from meshy.views import open_map_picker

    dev = window.device_info
    lat = dev.latitude if dev and dev.latitude is not None else 0.0
    lon = dev.longitude if dev and dev.longitude is not None else 0.0
    open_map_picker(
        window,
        _("Share Location"),
        _("Share"),
        lat,
        lon,
        lambda la, lo: send_callback(f"{la:.6f},{lo:.6f}"),
    )



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""QR code scanner using XDG Camera Portal + PipeWire + GStreamer."""

import logging
import os

try:
    from pyzbar import pyzbar as _pyzbar
except ImportError:
    _pyzbar = None

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Gst", "1.0")
from gi.repository import Adw, Gio, GLib, Gst, Gtk

log = logging.getLogger(__name__)

Gst.init(None)

PORTAL_BUS_NAME = "org.freedesktop.portal.Desktop"
PORTAL_OBJECT_PATH = "/org/freedesktop/portal/desktop"
PORTAL_CAMERA_IFACE = "org.freedesktop.portal.Camera"
PORTAL_REQUEST_IFACE = "org.freedesktop.portal.Request"


class QRScannerDialog:
    """Dialog that opens the camera via XDG portal and scans for QR codes."""

    def __init__(self, window, on_result):
        self._window = window
        self._on_result = on_result
        self._pipeline = None
        self._dialog = None
        self._signal_id = None

    def start(self):
        """Request camera access via the XDG portal."""
        try:
            self._bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        except GLib.Error as e:
            log.error(f"Cannot connect to session bus: {e.message}")
            self._show_error(_("Cannot connect to session bus: {}").format(e.message))
            return

        # Check if camera portal is available
        try:
            result = self._bus.call_sync(
                PORTAL_BUS_NAME,
                PORTAL_OBJECT_PATH,
                "org.freedesktop.DBus.Properties",
                "Get",
                GLib.Variant("(ss)", (PORTAL_CAMERA_IFACE, "IsCameraPresent")),
                GLib.VariantType.new("(v)"),
                Gio.DBusCallFlags.NONE,
                5000,
                None,
            )
            is_present = result.get_child_value(0).get_variant().get_boolean()
            if not is_present:
                self._show_error(_("No camera detected on this system."))
                return
        except GLib.Error as e:
            log.warning(f"Could not check IsCameraPresent: {e.message}")
            # Continue anyway, the portal might still work

        self._access_camera()

    def _access_camera(self):
        """Call AccessCamera on the portal."""
        token = f"meshy_{os.getpid()}"
        sender = self._bus.get_unique_name().replace(".", "_").lstrip(":")
        request_path = f"/org/freedesktop/portal/desktop/request/{sender}/{token}"

        log.info(f"AccessCamera: subscribing to {request_path}")

        # Subscribe to the Response signal before making the call
        self._signal_id = self._bus.signal_subscribe(
            PORTAL_BUS_NAME,
            PORTAL_REQUEST_IFACE,
            "Response",
            request_path,
            None,
            Gio.DBusSignalFlags.NO_MATCH_RULE,
            self._on_access_response,
        )

        params = GLib.Variant(
            "(a{sv})",
            (
                {
                    "handle_token": GLib.Variant("s", token),
                },
            ),
        )

        self._bus.call(
            PORTAL_BUS_NAME,
            PORTAL_OBJECT_PATH,
            PORTAL_CAMERA_IFACE,
            "AccessCamera",
            params,
            None,  # accept any reply type
            Gio.DBusCallFlags.NONE,
            30000,
            None,
            self._on_access_call_done,
        )

    def _on_access_call_done(self, bus, result):
        try:
            res = bus.call_finish(result)
            log.info(f"AccessCamera returned: {res}")
        except GLib.Error as e:
            log.error(f"AccessCamera call failed: {e.message}")
            self._cleanup_signal()
            self._show_error(_("Could not access camera: {}").format(e.message))

    def _on_access_response(self, bus, sender, path, iface, signal, params):
        log.info(f"AccessCamera response received: {params}")
        self._cleanup_signal()

        response = params.get_child_value(0).get_uint32()
        if response != 0:
            log.warning(f"Camera access denied, response={response}")
            GLib.idle_add(lambda: self._show_error(_("Camera access was denied.")) or False)
            return

        # Permission granted, open PipeWire remote
        GLib.idle_add(lambda: self._open_pipewire_remote() or False)

    def _cleanup_signal(self):
        if self._signal_id:
            self._bus.signal_unsubscribe(self._signal_id)
            self._signal_id = None

    def _open_pipewire_remote(self):
        """Get a PipeWire fd from the portal."""
        try:
            msg = Gio.DBusMessage.new_method_call(
                PORTAL_BUS_NAME,
                PORTAL_OBJECT_PATH,
                PORTAL_CAMERA_IFACE,
                "OpenPipeWireRemote",
            )
            msg.set_body(GLib.Variant("(a{sv})", ({},)))

            result = self._bus.send_message_with_reply_sync(
                msg,
                Gio.DBusSendMessageFlags.NONE,
                5000,
                None,
            )
            # send_message_with_reply_sync returns (reply_msg, out_serial)
            reply_msg = result[0] if isinstance(result, tuple) else result

            if reply_msg.get_message_type() == Gio.DBusMessageType.ERROR:
                error_name = reply_msg.get_error_name()
                log.error(f"OpenPipeWireRemote error: {error_name}")
                self._show_error(_("Could not open camera stream."))
                return

            fd_list = reply_msg.get_unix_fd_list()
            if not fd_list or fd_list.get_length() == 0:
                log.error("OpenPipeWireRemote: no fd in reply")
                self._show_error(_("No camera file descriptor received."))
                return

            pw_fd = fd_list.get(0)
            log.info(f"Got PipeWire fd: {pw_fd}")
            self._start_scanning(pw_fd)

        except GLib.Error as e:
            log.error(f"OpenPipeWireRemote failed: {e.message}")
            self._show_error(_("Could not open camera: {}").format(e.message))
        except Exception as e:
            log.error(f"OpenPipeWireRemote unexpected error: {e}")
            self._show_error(_("Could not open camera: {}").format(e))

    def _start_scanning(self, pw_fd):
        """Build GStreamer pipeline and show the camera dialog."""
        self._dialog = Adw.Dialog()
        self._dialog.set_title(_("Scan QR Code"))
        self._dialog.set_content_width(400)
        self._dialog.set_content_height(400)
        self._dialog.connect("closed", self._on_dialog_closed)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=8,
            margin_end=8,
            margin_bottom=8,
        )

        # Video preview
        self._picture = Gtk.Picture()
        self._picture.set_size_request(360, 270)
        self._picture.set_vexpand(True)
        self._picture.set_content_fit(Gtk.ContentFit.CONTAIN)
        content.append(self._picture)

        # Status label
        self._status_label = Gtk.Label(label=_("Point the camera at a QR code..."))
        self._status_label.add_css_class("dim-label")
        content.append(self._status_label)

        toolbar_view.set_content(content)
        self._dialog.set_child(toolbar_view)

        # Build GStreamer pipeline
        pipeline_str = (
            f"pipewiresrc fd={pw_fd} ! "
            "videoconvert ! "
            "video/x-raw,format=RGB ! "
            "tee name=t "
            "t. ! queue ! videoconvert ! gtk4paintablesink name=sink "
            "t. ! queue max-size-buffers=1 leaky=downstream ! "
            "videoscale ! video/x-raw,width=640,height=480 ! "
            "videoconvert ! video/x-raw,format=GRAY8 ! "
            "appsink name=qrsink emit-signals=true max-buffers=1 drop=true"
        )
        try:
            self._pipeline = Gst.parse_launch(pipeline_str)
        except GLib.Error as e:
            log.error(f"Failed to create GStreamer pipeline: {e.message}")
            self._show_error(_("Camera pipeline error: {}").format(e.message))
            return

        # Connect the paintable sink to the Picture widget
        sink = self._pipeline.get_by_name("sink")
        paintable = sink.get_property("paintable")
        self._picture.set_paintable(paintable)

        # Connect to appsink for QR decoding
        appsink = self._pipeline.get_by_name("qrsink")
        appsink.connect("new-sample", self._on_new_sample)

        # Watch for pipeline errors
        bus = self._pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", self._on_pipeline_error)

        self._pipeline.set_state(Gst.State.PLAYING)
        self._dialog.present(self._window)
        self._scan_active = True
        self._scan_start_time = GLib.get_monotonic_time()
        self._hint_shown = False
        self._timeout_shown = False
        self._hint_timer_id = GLib.timeout_add(8000, self._on_scan_hint)
        self._timeout_timer_id = GLib.timeout_add(30000, self._on_scan_timeout)

    def _on_new_sample(self, appsink):
        """Pull a frame from appsink and try to decode QR codes."""
        if not self._scan_active:
            return Gst.FlowReturn.OK

        sample = appsink.emit("pull-sample")
        if not sample:
            return Gst.FlowReturn.OK

        buf = sample.get_buffer()
        caps = sample.get_caps()
        struct = caps.get_structure(0)
        width = struct.get_value("width")
        height = struct.get_value("height")
        struct.get_string("format")

        success, mapinfo = buf.map(Gst.MapFlags.READ)
        if not success:
            return Gst.FlowReturn.OK

        if _pyzbar is None:
            buf.unmap(mapinfo)
            return Gst.FlowReturn.OK

        try:
            results = _pyzbar.decode(
                (bytes(mapinfo.data), width, height),
                symbols=[_pyzbar.ZBarSymbol.QRCODE],
            )
        except Exception as e:
            log.debug(f"QR decode error: {e}")
            buf.unmap(mapinfo)
            return Gst.FlowReturn.OK

        buf.unmap(mapinfo)

        for qr in results:
            data = qr.data.decode("utf-8", errors="ignore").strip()
            log.info(f"QR code detected: {data[:80]}...")
            self._scan_active = False
            GLib.idle_add(self._on_qr_found, data)
            return Gst.FlowReturn.OK

        return Gst.FlowReturn.OK

    def _on_scan_hint(self):
        """Show a hint after 8 seconds of no detection."""
        self._hint_timer_id = None
        if self._scan_active:
            self._status_label.set_label(
                _(
                    "No QR code detected yet.\n"
                    "Make sure the code is well-lit, sharp, and fills most of the frame."
                )
            )
        return False

    def _on_scan_timeout(self):
        """Show a give-up message after 30 seconds."""
        self._timeout_timer_id = None
        if self._scan_active:
            self._status_label.set_label(
                _("Could not detect QR code.\n" "Try importing the contact via clipboard instead.")
            )
        return False

    def _on_qr_found(self, data):
        """Handle a detected meshcore:// QR code."""
        self._status_label.set_label(_("QR code found!"))
        self._stop_pipeline()
        if self._dialog:
            self._dialog.close()
        self._on_result(data)
        return False

    def _on_pipeline_error(self, bus, msg):
        err, debug = msg.parse_error()
        log.error(f"GStreamer error: {err.message} ({debug})")
        GLib.idle_add(lambda: self._show_error(_("Camera error: {}").format(err.message)) or False)

    def _on_dialog_closed(self, dialog):
        self._stop_pipeline()

    def _stop_pipeline(self):
        self._scan_active = False
        if getattr(self, "_hint_timer_id", None):
            GLib.source_remove(self._hint_timer_id)
            self._hint_timer_id = None
        if getattr(self, "_timeout_timer_id", None):
            GLib.source_remove(self._timeout_timer_id)
            self._timeout_timer_id = None
        if self._pipeline:
            self._pipeline.set_state(Gst.State.NULL)
            self._pipeline = None

    def _show_error(self, message):
        dialog = Adw.AlertDialog(heading=_("Camera Error"), body=message)
        dialog.add_response("ok", _("OK"))
        dialog.present(self._window)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""macOS application activation helper.

GTK4 on macOS needs explicit NSApplication activation to show windows properly.
"""

def activate_macos_app():
    """Activate the macOS application so windows appear.

    This configures NSApplication to behave as a regular GUI app and
    brings it to the foreground.
    """
    try:
        from AppKit import (
            NSApplication,
            NSApplicationActivationPolicyRegular,
            NSApplicationActivateAllWindows,
            NSApplicationActivateIgnoringOtherApps,
        )

        # Get the shared NSApplication instance
        app = NSApplication.sharedApplication()

        # Set activation policy to regular app (shows in Dock, can be focused)
        # This must be done before activating
        app.setActivationPolicy_(NSApplicationActivationPolicyRegular)

        # Activate the app and bring it to the foreground
        # activateIgnoringOtherApps: brings app to front regardless of other apps
        app.activateIgnoringOtherApps_(True)

        import logging
        log = logging.getLogger(__name__)
        log.info("macOS app activated successfully")

        return True
    except ImportError:
        # PyObjC not available
        import logging
        log = logging.getLogger(__name__)
        log.warning("PyObjC AppKit not available - window may not activate automatically")
        return False
    except Exception as e:
        import logging
        log = logging.getLogger(__name__)
        log.warning(f"Failed to activate macOS app: {e}")
        import traceback
        traceback.print_exc()
        return False



"""Smoke-test that all meshy modules can be imported.

Catches missing symbols, broken re-exports, and circular imports
that would otherwise only surface at runtime.
"""

import importlib
import pkgutil

import meshy

# Modules that need optional GI typelibs (Shumate, WebKit, etc.)
# or platform-specific frameworks (PyObjC) — skip when unavailable.
_OPTIONAL_DEPS = {"Shumate", "WebKit", "CoreBluetooth", "CoreLocation", "ServiceManagement", "objc"}


def _is_optional_dep_error(exc):
    msg = str(exc)
    if any(dep in msg for dep in _OPTIONAL_DEPS):
        return True
    if "g-resource-error-quark" in msg:
        return True
    return False


def _iter_meshy_modules():
    """Yield dotted names of all meshy submodules (recursive)."""
    prefix = meshy.__name__ + "."
    for info in pkgutil.walk_packages(meshy.__path__, prefix, onerror=lambda _name: None):
        yield info.name


def test_all_modules_importable():
    failures = {}
    for name in _iter_meshy_modules():
        try:
            importlib.import_module(name)
        except Exception as exc:
            if _is_optional_dep_error(exc):
                continue
            failures[name] = exc

    if failures:
        lines = [f"  {mod}: {err}" for mod, err in sorted(failures.items())]
        raise AssertionError(f"{len(failures)} module(s) failed to import:\n" + "\n".join(lines))



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""USB Serial transport for MeshCore companion devices.

Uses pyserial to communicate with MeshCore devices over USB serial ports.
"""

import contextlib
import glob
import logging
import time

from gi.repository import GLib
from meshy.transport_base import ConnectionState, FrameTransport

log = logging.getLogger(__name__)

SERIAL_BAUD_RATE = 115200


class UsbDevice:
    """Represents a discovered USB serial device."""

    def __init__(self, path: str, name: str = ""):
        self.path = path
        self.address = path  # Use path as address for storage key
        self.name = name or path

    def __repr__(self):
        return f"UsbDevice({self.path})"


class UsbSerialTransport(FrameTransport):
    """USB serial transport using pyserial."""

    _LABEL = "USB"

    def __init__(self):
        super().__init__()
        self._port = None
        self._connection_error: str | None = None
        self._is_permission_error: bool = False

    def get_serial_devices(self) -> list[UsbDevice]:
        """Scan for available USB serial devices."""
        devices = []

        from meshy.platforms import is_macos
        if is_macos():
            patterns = ["/dev/cu.usbserial-*", "/dev/cu.usbmodem-*"]
        else:
            patterns = ["/dev/ttyACM*", "/dev/ttyUSB*"]

        for pattern in patterns:
            for path in sorted(glob.glob(pattern)):
                name = path.split("/")[-1]
                devices.append(UsbDevice(path=path, name=name))
        return devices

    def get_paired_devices(self) -> list[UsbDevice]:
        """Return available serial devices (USB doesn't have pairing)."""
        return self.get_serial_devices()

    def connect_device(self, device_path: str):
        """Connect to a USB serial device."""
        import serial

        if self._state == ConnectionState.CONNECTED:
            self.disconnect()

        self._set_state(ConnectionState.CONNECTING)
        self._connected_address = device_path

        try:
            self._port = serial.Serial(
                device_path,
                baudrate=SERIAL_BAUD_RATE,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                xonxoff=False,
                rtscts=False,
                timeout=None,
            )
            self._port.rts = False
            self._port.reset_input_buffer()
            self._port.reset_output_buffer()
            log.info(f"Opened serial port {device_path} at {SERIAL_BAUD_RATE} baud")

            self._start_read_thread()
            self._set_state(ConnectionState.CONNECTED)

        except Exception as e:
            error_msg = str(e)
            log.error(f"Failed to open serial port {device_path}: {error_msg}")
            self._connection_error = error_msg
            self._is_permission_error = "Permission denied" in error_msg or "Errno 13" in error_msg
            self._connected_address = None
            self._set_state(ConnectionState.DISCONNECTED)

    def disconnect(self):
        """Disconnect from the serial device."""
        if self._state in (ConnectionState.DISCONNECTED, ConnectionState.DISCONNECTING):
            return

        self._set_state(ConnectionState.DISCONNECTING)
        self._running = False
        self._close_io()
        self._connected_address = None
        self._set_state(ConnectionState.DISCONNECTED)

    # ─── FrameTransport implementation ────────────────────────

    def _is_io_ready(self) -> bool:
        return self._port is not None

    def _write_bytes(self, frame: bytes):
        self._port.write(frame)

    def _close_io(self):
        if self._port:
            with contextlib.suppress(Exception):
                self._port.close()
            self._port = None

    def _read_loop(self, gen: int):
        """Background thread that reads from the serial port."""
        port = self._port
        while self._running and gen == self._connect_gen and port:
            try:
                waiting = port.in_waiting
                if waiting > 0:
                    data = port.read(waiting)
                    if data:
                        self._rx_buffer.extend(data)
                        self._process_rx_buffer()
                else:
                    time.sleep(0.05)
            except OSError as e:
                if self._running and gen == self._connect_gen:
                    log.error(f"USB read error: {e}")
                    GLib.idle_add(self._handle_disconnect, gen)
                break
            except Exception as e:
                if self._running and gen == self._connect_gen:
                    log.error(f"USB read error: {e}")
                break



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""CoreBluetooth BLE transport for MeshCore companion devices (macOS).

Uses the macOS CoreBluetooth framework to discover, connect, and communicate with
MeshCore devices via the Nordic UART Service (NUS).

This implementation uses PyObjC to interface with macOS CoreBluetooth APIs.
"""

import logging

from meshy.platforms.ble_base import (
    BleDevice,
    NUS_RX_CHAR_UUID,
    NUS_SERVICE_UUID,
    NUS_TX_CHAR_UUID,
)
from meshy.transport_base import ConnectionState

log = logging.getLogger(__name__)

# BLE connection timeouts and retry settings
_BLE_CONNECTION_TIMEOUT_S = 30

# Import CoreBluetooth framework via PyObjC
try:
    from Foundation import NSObject
    from CoreBluetooth import (
        CBUUID,
        CBCentralManager,
        CBCharacteristicWriteWithResponse,
        CBManagerStatePoweredOff,
        CBManagerStatePoweredOn,
        CBManagerStateUnauthorized,
        CBManagerStateUnknown,
        CBManagerStateUnsupported,
        CBPeripheral,
    )

    COREBLUETOOTH_AVAILABLE = True

except ImportError:
    log.warning("PyObjC CoreBluetooth not available - BLE will not work on macOS")
    COREBLUETOOTH_AVAILABLE = False
    NSObject = object  # Fallback for class definition


class BleTransportCoreBluetooth(NSObject):
    """BLE transport using CoreBluetooth framework (macOS).

    This class duck-types the BleTransportBase interface and extends NSObject
    to serve as a CoreBluetooth delegate. It does not inherit from BleTransportBase
    to avoid metaclass conflicts with NSObject.
    """

    def init(self):
        """Initialize the CoreBluetooth transport (Objective-C style required for PyObjC)."""
        if not COREBLUETOOTH_AVAILABLE:
            return None

        # Initialize parent NSObject using objc.super
        from objc import super as objc_super
        self = objc_super(BleTransportCoreBluetooth, self).init()
        if self is None:
            return None

        # Initialize instance variables
        self._state = ConnectionState.DISCONNECTED
        self._central_manager = None
        self._connected_peripheral = None
        self._discovered_devices = {}
        self._device_address = None
        self._tx_characteristic = None
        self._rx_characteristic = None
        self._connection_timeout_id = None  # GLib timeout source ID

        # Callbacks (from BleTransportBase interface)
        self.on_state_changed = None
        self.on_data_received = None
        self.on_device_discovered = None
        self.on_passkey_requested = None

        # Initialize CoreBluetooth central manager on main thread
        from gi.repository import GLib
        GLib.idle_add(self._init_central_manager)

        log.info("CoreBluetooth transport initialized")
        return self

    def _init_central_manager(self):
        """Initialize the CBCentralManager (must be called on main thread).

        Note: This will crash if called from a command-line Python process
        without a proper Info.plist containing NSBluetoothAlwaysUsageDescription.
        macOS's TCC (privacy system) requires this before any Bluetooth access.
        """
        if not COREBLUETOOTH_AVAILABLE:
            return False

        # Check if we're running in a proper app bundle with Info.plist
        # If not, skip BLE initialization to avoid TCC crash
        try:
            from Foundation import NSBundle, NSURL
            import sys
            import os

            # Try to find the .app bundle by walking up from the executable
            bundle = None
            executable_path = os.path.realpath(sys.argv[0])
            log.debug(f"Executable path: {executable_path}")

            # Check if we're inside a .app bundle
            # Look for any .app/Contents pattern (could be MacOS, Resources, etc.)
            if '.app/Contents/' in executable_path:
                # Extract the .app bundle path - find first .app and include it
                parts = executable_path.split('.app/Contents/')
                if len(parts) >= 2:
                    app_path = parts[0] + '.app'
                    log.debug(f"Detected .app bundle path: {app_path}")
                    if os.path.exists(app_path):
                        bundle_url = NSURL.fileURLWithPath_(app_path)
                        bundle = NSBundle.bundleWithURL_(bundle_url)
                        log.debug(f"Loaded bundle from: {app_path}")
                    else:
                        log.warning(f"App path does not exist: {app_path}")

            # Fallback to main bundle
            if not bundle:
                bundle = NSBundle.mainBundle()
                log.debug(f"Using main bundle: {bundle.bundlePath()}")

            info = bundle.infoDictionary()
            if info:
                has_bluetooth_key = "NSBluetoothAlwaysUsageDescription" in info
                log.debug(f"Bundle has Info.plist: YES, has Bluetooth key: {has_bluetooth_key}")
                if has_bluetooth_key:
                    log.debug(f"Bluetooth description: {info['NSBluetoothAlwaysUsageDescription']}")
            else:
                log.warning("Bundle has no Info.plist")

            # Check if NSBluetoothAlwaysUsageDescription exists in Info.plist
            if info and "NSBluetoothAlwaysUsageDescription" in info:
                log.info("Info.plist contains Bluetooth permissions, initializing CoreBluetooth")
                self._central_manager = CBCentralManager.alloc().initWithDelegate_queue_(self, None)
                log.info("CoreBluetooth initialized successfully")
            else:
                log.warning("No Info.plist with NSBluetoothAlwaysUsageDescription found")
                log.warning("BLE disabled - run from .app bundle for Bluetooth support")
                log.info("USB and TCP connectivity still available")
                # Don't initialize - would cause TCC crash
        except Exception as e:
            log.error(f"Failed to check for Info.plist: {e}", exc_info=True)
            log.warning("BLE disabled to prevent TCC crash")

        return False  # Don't repeat

    # ─── Properties (BleTransportBase interface) ───────────────────

    @property
    def state(self) -> ConnectionState:
        return self._state

    @property
    def is_connected(self) -> bool:
        return self._state == ConnectionState.CONNECTED

    @property
    def connected_address(self) -> str | None:
        return self._device_address

    @property
    def discovered_devices(self) -> list[BleDevice]:
        return list(self._discovered_devices.values())

    # ─── BLE Operations (BleTransportBase interface) ───────────────

    def start_scan(self):
        """Start scanning for BLE devices."""
        if not COREBLUETOOTH_AVAILABLE:
            log.error("CoreBluetooth not available")
            return

        if not self.is_bluetooth_enabled():
            log.error("Bluetooth is not enabled")
            return

        log.info("Starting BLE scan...")
        self._set_state(ConnectionState.SCANNING)
        self._discovered_devices.clear()

        if self._central_manager:
            # Scan for devices advertising the NUS service
            service_uuids = [CBUUID.UUIDWithString_(NUS_SERVICE_UUID)]
            self._central_manager.scanForPeripheralsWithServices_options_(service_uuids, None)

    def stop_scan(self):
        """Stop scanning for BLE devices."""
        if self._central_manager:
            self._central_manager.stopScan()
        self._set_state(ConnectionState.DISCONNECTED)

    def connect_device(self, device_path: str):
        """Connect to a device by its path/identifier.

        device_path is the peripheral UUID string.
        """
        if not COREBLUETOOTH_AVAILABLE:
            log.error("CoreBluetooth not available")
            return

        # Find the peripheral in discovered devices
        peripheral = None
        for device in self._discovered_devices.values():
            if device.path == device_path:
                peripheral = device._peripheral  # Internal reference
                break

        # If not found in discovered devices, try to retrieve by UUID
        # (this handles paired devices that haven't been scanned this session)
        if not peripheral:
            log.info(f"Device not in discovered list, retrieving by UUID: {device_path}")
            try:
                from Foundation import NSUUID
                uuid = NSUUID.alloc().initWithUUIDString_(device_path)
                if uuid and self._central_manager:
                    peripherals = self._central_manager.retrievePeripheralsWithIdentifiers_([uuid])
                    if peripherals and len(peripherals) > 0:
                        peripheral = peripherals[0]
                        log.info(f"Retrieved peripheral: {peripheral.name()}")
                    else:
                        log.error(f"Could not retrieve peripheral with UUID: {device_path}")
                        return
            except Exception as e:
                log.error(f"Failed to retrieve peripheral: {e}")
                return

        if not peripheral:
            log.error(f"Device {device_path} not found")
            return

        log.info(f"Connecting to device {device_path}...")
        self._set_state(ConnectionState.CONNECTING)
        self._device_address = device_path

        if self._central_manager:
            self._central_manager.connectPeripheral_options_(peripheral, None)

            # Start connection timeout
            from gi.repository import GLib
            self._cancel_connection_timeout()  # Cancel any existing timeout
            self._connection_timeout_id = GLib.timeout_add_seconds(
                _BLE_CONNECTION_TIMEOUT_S,
                self._on_connection_timeout
            )
            log.info(f"Connection timeout scheduled ({_BLE_CONNECTION_TIMEOUT_S} seconds)")

    def connect_by_address(self, address: str):
        """Connect to a device by its MAC address.

        On macOS, we use UUID instead of MAC address.
        This attempts to retrieve a known peripheral by UUID.
        """
        if not COREBLUETOOTH_AVAILABLE:
            log.error("CoreBluetooth not available")
            return

        if not self._central_manager:
            log.warning("CoreBluetooth manager not initialized yet - cannot connect by address")
            return

        try:
            from Foundation import NSUUID
            uuid = NSUUID.alloc().initWithUUIDString_(address)
            if not uuid:
                log.error(f"Invalid UUID: {address}")
                return

            # Retrieve peripheral by identifier
            peripherals = self._central_manager.retrievePeripheralsWithIdentifiers_([uuid])
            if not peripherals or len(peripherals) == 0:
                log.error(f"No peripheral found with UUID {address}")
                return

            peripheral = peripherals[0]
            log.info(f"Connecting to peripheral {peripheral.name()}...")
            self._set_state(ConnectionState.CONNECTING)
            self._device_address = address
            self._central_manager.connectPeripheral_options_(peripheral, None)

            # Start connection timeout
            from gi.repository import GLib
            self._cancel_connection_timeout()  # Cancel any existing timeout
            self._connection_timeout_id = GLib.timeout_add_seconds(
                _BLE_CONNECTION_TIMEOUT_S,
                self._on_connection_timeout
            )
            log.info(f"Connection timeout scheduled ({_BLE_CONNECTION_TIMEOUT_S} seconds)")

        except Exception as e:
            log.error(f"Failed to connect by address: {e}")

    def pair_device(self, device_path: str):
        """Pair with a device (handled automatically by macOS)."""
        # On macOS, pairing is handled automatically by the system
        # Just connect to the device
        self.connect_device(device_path)

    def disconnect(self):
        """Disconnect from the current device."""
        self._cancel_connection_timeout()  # Cancel any pending timeout
        if self._connected_peripheral and self._central_manager:
            self._central_manager.cancelPeripheralConnection_(self._connected_peripheral)
        self._set_state(ConnectionState.DISCONNECTED)

    def _cancel_connection_timeout(self):
        """Cancel any pending connection timeout."""
        if self._connection_timeout_id is not None:
            from gi.repository import GLib
            GLib.source_remove(self._connection_timeout_id)
            self._connection_timeout_id = None
            log.debug("Connection timeout cancelled")

    def _on_connection_timeout(self):
        """Handle connection timeout."""
        if self._state == ConnectionState.CONNECTING:
            log.error(f"BLE connection timeout after {_BLE_CONNECTION_TIMEOUT_S} seconds")
            # Emit error to UI
            if self.on_state_changed:
                from gi.repository import GLib
                GLib.idle_add(self._emit_error, "Connection timeout - device not responding")
            # Force disconnect
            self.disconnect()
        self._connection_timeout_id = None
        return False  # Don't repeat timeout

    def _emit_error(self, message: str):
        """Emit error message via state changed callback."""
        # This is a workaround since BleTransportBase has on_error callback
        # but we're duck-typing the interface without inheriting
        log.error(f"Emitting error: {message}")
        if self.on_state_changed:
            self.on_state_changed(ConnectionState.DISCONNECTED)
        return False

    def send_data(self, data: bytes):
        """Send data to the connected device."""
        if not self.is_connected or not self._rx_characteristic:
            log.warning("Cannot send data: not connected")
            return

        if not self._connected_peripheral:
            log.warning("No connected peripheral")
            return

        # Convert bytes to NSData
        from Foundation import NSData
        ns_data = NSData.dataWithBytes_length_(data, len(data))

        # Write to RX characteristic (app -> device)
        self._connected_peripheral.writeValue_forCharacteristic_type_(
            ns_data,
            self._rx_characteristic,
            CBCharacteristicWriteWithResponse
        )
        log.debug(f"Sent {len(data)} bytes to device")

    def is_bluetooth_enabled(self) -> bool:
        """Check if Bluetooth is enabled."""
        if not self._central_manager:
            return False
        state = self._central_manager.state()
        return state == CBManagerStatePoweredOn

    def has_bluetooth_adapter(self) -> bool:
        """Check if a Bluetooth adapter is available."""
        if not self._central_manager:
            return False
        state = self._central_manager.state()
        return state != CBManagerStateUnsupported and state != CBManagerStateUnknown

    def get_paired_devices(self) -> list[BleDevice]:
        """Get list of paired devices that support NUS.

        Retrieves devices that have been successfully paired by checking:
        1. Saved device UUIDs from GSettings
        2. Currently connected peripherals with NUS service
        """
        if not self._central_manager:
            return []

        devices = []

        # Get saved paired device UUIDs from GSettings
        try:
            from gi.repository import Gio
            settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
            saved_uuids = settings.get_value("ble-paired-devices")

            if saved_uuids and len(saved_uuids) > 0:
                from Foundation import NSUUID

                # Convert string UUIDs to NSUUID objects
                uuid_objects = []
                for uuid_str in saved_uuids:
                    uuid_obj = NSUUID.alloc().initWithUUIDString_(uuid_str)
                    if uuid_obj:
                        uuid_objects.append(uuid_obj)

                # Retrieve peripherals by their identifiers
                if uuid_objects:
                    peripherals = self._central_manager.retrievePeripheralsWithIdentifiers_(uuid_objects)

                    for peripheral in peripherals:
                        uuid = peripheral.identifier().UUIDString()
                        name = peripheral.name() or "Unknown Device"

                        device = BleDevice(
                            path=uuid,
                            address=uuid,
                            name=name
                        )
                        device._peripheral = peripheral
                        devices.append(device)

        except Exception as e:
            log.warning(f"Failed to retrieve saved paired devices: {e}")

        # Also check for currently connected peripherals with NUS service
        try:
            service_uuids = [CBUUID.UUIDWithString_(NUS_SERVICE_UUID)]
            connected = self._central_manager.retrieveConnectedPeripheralsWithServices_(service_uuids)

            # Add connected devices if not already in list
            existing_uuids = {d.path for d in devices}
            for peripheral in connected:
                uuid = peripheral.identifier().UUIDString()
                if uuid not in existing_uuids:
                    name = peripheral.name() or "Unknown Device"
                    device = BleDevice(
                        path=uuid,
                        address=uuid,
                        name=name
                    )
                    device._peripheral = peripheral
                    devices.append(device)

        except Exception as e:
            log.warning(f"Failed to retrieve connected peripherals: {e}")

        return devices

    def remove_device(self, device_path: str) -> bool:
        """Remove/unpair a device (managed by system on macOS)."""
        # macOS manages pairing at the system level
        # We can only disconnect, not unpair
        log.info("Device pairing is managed by macOS System Settings")
        return False

    # ─── CoreBluetooth Delegate Methods ────────────────────────────

    def centralManagerDidUpdateState_(self, central):
        """Called when the central manager state changes."""
        state = central.state()
        log.info(f"CoreBluetooth state changed: {state}")

        if state == CBManagerStatePoweredOn:
            log.info("Bluetooth is powered on and ready")
            if self._state == ConnectionState.SCANNING:
                # Resume scanning if we were scanning before
                pass
        elif state == CBManagerStatePoweredOff:
            log.warning("Bluetooth is powered off")
        elif state == CBManagerStateUnauthorized:
            log.error("Bluetooth access not authorized")

    def centralManager_didDiscoverPeripheral_advertisementData_RSSI_(
        self, central, peripheral, advertisement_data, rssi
    ):
        """Called when a peripheral is discovered during scanning."""
        # Get peripheral UUID as identifier
        uuid = peripheral.identifier().UUIDString()
        name = peripheral.name() or "Unknown"

        log.info(f"Discovered peripheral: {name} ({uuid}), RSSI: {rssi}")

        # Create BleDevice
        device = BleDevice(
            path=uuid,
            address=uuid,  # macOS uses UUID, not MAC address
            name=name
        )
        # Store reference to peripheral for connection
        device._peripheral = peripheral

        self._discovered_devices[uuid] = device

        # Notify callback
        if self.on_device_discovered:
            from gi.repository import GLib
            GLib.idle_add(lambda: self.on_device_discovered(device) if self.on_device_discovered else None)

    def centralManager_didConnectPeripheral_(self, central, peripheral):
        """Called when a peripheral is successfully connected."""
        self._cancel_connection_timeout()  # Cancel timeout on successful connection
        log.info(f"Connected to peripheral: {peripheral.name()}")
        self._connected_peripheral = peripheral
        peripheral.setDelegate_(self)

        # Discover NUS service
        service_uuids = [CBUUID.UUIDWithString_(NUS_SERVICE_UUID)]
        peripheral.discoverServices_(service_uuids)

    def centralManager_didFailToConnectPeripheral_error_(self, central, peripheral, error):
        """Called when connection to a peripheral fails."""
        self._cancel_connection_timeout()  # Cancel timeout on connection failure
        log.error(f"Failed to connect to peripheral: {error}")
        self._set_state(ConnectionState.DISCONNECTED)

    def centralManager_didDisconnectPeripheral_error_(self, central, peripheral, error):
        """Called when a peripheral disconnects."""
        log.info(f"Peripheral disconnected: {error if error else 'user initiated'}")
        self._connected_peripheral = None
        self._tx_characteristic = None
        self._rx_characteristic = None
        self._set_state(ConnectionState.DISCONNECTED)

    # ─── CBPeripheralDelegate Methods ──────────────────────────────

    def peripheral_didDiscoverServices_(self, peripheral, error):
        """Called when services are discovered."""
        if error:
            log.error(f"Service discovery error: {error}")
            self._set_state(ConnectionState.DISCONNECTED)
            return

        # Find the NUS service
        nus_uuid = CBUUID.UUIDWithString_(NUS_SERVICE_UUID)
        nus_service = None

        for service in peripheral.services():
            if service.UUID().isEqual_(nus_uuid):
                nus_service = service
                break

        if not nus_service:
            log.error("NUS service not found on peripheral")
            self._set_state(ConnectionState.DISCONNECTED)
            self._central_manager.cancelPeripheralConnection_(peripheral)
            return

        log.info("NUS service found, discovering characteristics...")
        # Discover TX and RX characteristics
        char_uuids = [
            CBUUID.UUIDWithString_(NUS_TX_CHAR_UUID),
            CBUUID.UUIDWithString_(NUS_RX_CHAR_UUID)
        ]
        peripheral.discoverCharacteristics_forService_(char_uuids, nus_service)

    def peripheral_didDiscoverCharacteristicsForService_error_(self, peripheral, service, error):
        """Called when characteristics are discovered."""
        if error:
            log.error(f"Characteristic discovery error: {error}")
            self._set_state(ConnectionState.DISCONNECTED)
            return

        tx_uuid = CBUUID.UUIDWithString_(NUS_TX_CHAR_UUID)
        rx_uuid = CBUUID.UUIDWithString_(NUS_RX_CHAR_UUID)

        for characteristic in service.characteristics():
            if characteristic.UUID().isEqual_(tx_uuid):
                self._tx_characteristic = characteristic
                # Enable notifications for TX (device -> app)
                peripheral.setNotifyValue_forCharacteristic_(True, characteristic)
                log.info("Enabled notifications for TX characteristic")

            elif characteristic.UUID().isEqual_(rx_uuid):
                self._rx_characteristic = characteristic
                log.info("Found RX characteristic")

        # Check if we have both characteristics
        if self._tx_characteristic and self._rx_characteristic:
            log.info("NUS service ready, connection established")

            # Save this device as paired
            self._save_paired_device(peripheral)

            self._set_state(ConnectionState.CONNECTED)
        else:
            log.error("Failed to find required NUS characteristics")
            self._set_state(ConnectionState.DISCONNECTED)
            self._central_manager.cancelPeripheralConnection_(peripheral)

    def peripheral_didUpdateValueForCharacteristic_error_(self, peripheral, characteristic, error):
        """Called when a characteristic value is updated (notifications)."""
        if error:
            log.error(f"Characteristic update error: {error}")
            return

        # Check if this is the TX characteristic
        tx_uuid = CBUUID.UUIDWithString_(NUS_TX_CHAR_UUID)
        if not characteristic.UUID().isEqual_(tx_uuid):
            return

        # Read data from characteristic
        ns_data = characteristic.value()
        if not ns_data:
            return

        # Convert NSData to bytes
        data = bytes(ns_data.bytes()[:ns_data.length()])
        log.debug(f"Received {len(data)} bytes from device")

        # Notify callback on main thread
        if self.on_data_received:
            from gi.repository import GLib
            GLib.idle_add(lambda: self.on_data_received(data) if self.on_data_received else None)

    # ─── Helper Methods ─────────────────────────────────────────────

    def _set_state(self, state: ConnectionState):
        """Update connection state and notify callback."""
        if self._state != state:
            old = self._state
            self._state = state
            log.info(f"BLE state: {old.name} -> {state.name}")
            if self.on_state_changed:
                # Call on main thread via GLib
                from gi.repository import GLib

                GLib.idle_add(lambda: self.on_state_changed(state) if self.on_state_changed else None)

    def _save_paired_device(self, peripheral):
        """Save a successfully paired device UUID to persistent storage."""
        try:
            from gi.repository import Gio, GLib

            uuid = peripheral.identifier().UUIDString()
            settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")

            # Get existing paired devices
            saved_uuids = list(settings.get_value("ble-paired-devices"))

            # Add this device if not already saved
            if uuid not in saved_uuids:
                saved_uuids.append(uuid)
                settings.set_value("ble-paired-devices", GLib.Variant('as', saved_uuids))
                log.info(f"Saved paired device: {uuid}")
            else:
                log.debug(f"Device already in paired list: {uuid}")

        except Exception as e:
            log.error(f"Failed to save paired device: {e}")



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared base for frame-oriented MeshCore transports (USB, TCP).

Frame format: start_marker(1) + length_le(2) + payload(N)
  Host→Device: 0x3C + length + payload
  Device→Host: 0x3E + length + payload
"""

import logging
import threading
from collections.abc import Callable
from enum import IntEnum, auto

from gi.repository import GLib

log = logging.getLogger(__name__)

TX_MARKER = 0x3C  # Host → Device
RX_MARKER = 0x3E  # Device → Host
HEADER_SIZE = 3
MAX_RX_PAYLOAD = 176  # accept frames up to 176 bytes (firmware 1.16+)
MAX_TX_PAYLOAD = 172  # send at most 172 bytes for backward compat with older firmware


class ConnectionState(IntEnum):
    DISCONNECTED = auto()
    SCANNING = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    DISCONNECTING = auto()


class FrameTransport:
    """Base class for frame-oriented transports (USB serial, TCP).

    Subclasses must implement:
      - _write_bytes(frame: bytes)  — send raw bytes over the wire
      - _close_io()                 — close the underlying port/socket
      - _read_loop(gen: int)        — background read thread body
      - connect_device(path: str)   — establish the connection
    """

    # Subclass label for log messages
    _LABEL = "Transport"
    # Set of valid RX markers — override in TCP to also accept 0x3C
    _RX_MARKERS = frozenset((RX_MARKER,))

    def __init__(self):
        self._state = ConnectionState.DISCONNECTED
        self._connected_address: str | None = None
        self._read_thread: threading.Thread | None = None
        self._running = False
        self._rx_buffer = bytearray()
        self._connect_gen = 0

        # Callbacks — same interface as BleTransport
        self.on_state_changed: Callable | None = None
        self.on_data_received: Callable[[bytes], None] | None = None
        self.on_device_discovered: Callable | None = None
        self.on_passkey_requested: Callable | None = None

    # ─── Properties ───────────────────────────────────────────

    @property
    def state(self):
        return self._state

    @property
    def is_connected(self) -> bool:
        return self._state == ConnectionState.CONNECTED

    @property
    def connected_address(self) -> str | None:
        return self._connected_address

    @property
    def discovered_devices(self) -> list:
        return []

    # ─── State management ─────────────────────────────────────

    def _set_state(self, state):
        if self._state != state:
            old = self._state
            self._state = state
            log.info(
                f"{self._LABEL} state: {ConnectionState(old).name} -> "
                f"{ConnectionState(state).name}"
            )
            if self.on_state_changed:
                self.on_state_changed(state)

    # ─── Sending ──────────────────────────────────────────────

    def send_data(self, data: bytes):
        """Send a protocol frame wrapped in the wire framing."""
        if not self._is_io_ready() or self._state != ConnectionState.CONNECTED:
            log.warning(f"{self._LABEL}: cannot send, state={self._state}")
            return

        if len(data) > MAX_TX_PAYLOAD:
            log.error(f"Frame too large: {len(data)} > {MAX_TX_PAYLOAD}")
            return

        frame = bytearray(HEADER_SIZE + len(data))
        frame[0] = TX_MARKER
        frame[1] = len(data) & 0xFF
        frame[2] = (len(data) >> 8) & 0xFF
        frame[3:] = data

        try:
            self._write_bytes(bytes(frame))
            log.debug(f"{self._LABEL} TX [{len(data)} bytes]: {data[:20].hex()}...")
        except Exception as e:
            log.error(f"{self._LABEL} write failed: {e}")
            GLib.idle_add(self._handle_disconnect)

    # ─── Receiving ────────────────────────────────────────────

    def _process_rx_buffer(self):
        """Parse complete frames from the receive buffer."""
        while len(self._rx_buffer) >= HEADER_SIZE:
            marker = self._rx_buffer[0]
            if marker not in self._RX_MARKERS:
                idx = None
                for i, b in enumerate(self._rx_buffer):
                    if b in self._RX_MARKERS:
                        idx = i
                        break
                if idx is None:
                    self._rx_buffer.clear()
                    return
                log.debug(f"{self._LABEL}: skipping {idx} invalid bytes")
                del self._rx_buffer[:idx]
                continue

            payload_len = self._rx_buffer[1] | (self._rx_buffer[2] << 8)

            if payload_len > MAX_RX_PAYLOAD:
                log.debug(f"{self._LABEL}: invalid payload length {payload_len}, skipping")
                del self._rx_buffer[0]
                continue

            frame_size = HEADER_SIZE + payload_len
            if len(self._rx_buffer) < frame_size:
                return

            payload = bytes(self._rx_buffer[HEADER_SIZE:frame_size])
            del self._rx_buffer[:frame_size]

            log.debug(f"{self._LABEL} RX [{len(payload)} bytes]: {payload[:20].hex()}...")

            if self.on_data_received:
                GLib.idle_add(self._deliver_data, payload)

    def _deliver_data(self, data: bytes):
        """Deliver received data on the main thread."""
        if self.on_data_received:
            self.on_data_received(data)
        return False

    # ─── Disconnect handling ──────────────────────────────────

    def _handle_disconnect(self, gen: int = -1):
        """Handle unexpected disconnect. Ignore stale callbacks."""
        if gen != -1 and gen != self._connect_gen:
            log.debug(
                f"{self._LABEL}: ignoring stale disconnect " f"(gen {gen} != {self._connect_gen})"
            )
            return
        if self._state in (ConnectionState.DISCONNECTED, ConnectionState.DISCONNECTING):
            return
        log.info(f"{self._LABEL} disconnected")
        self._running = False
        self._close_io()
        self._connected_address = None
        self._set_state(ConnectionState.DISCONNECTED)

    def _start_read_thread(self):
        """Start the background read thread (call after IO is ready)."""
        self._running = True
        self._rx_buffer.clear()
        self._connect_gen += 1
        gen = self._connect_gen
        self._read_thread = threading.Thread(
            target=self._read_loop,
            args=(gen,),
            daemon=True,
        )
        self._read_thread.start()

    # ─── Stubs / no-ops ───────────────────────────────────────

    def connect_by_address(self, address: str):
        self.connect_device(address)

    def pair_device(self, device_path: str):
        self.connect_device(device_path)

    def start_scan(self):
        pass

    def stop_scan(self):
        pass

    # ─── Abstract (must be implemented by subclass) ───────────

    def _is_io_ready(self) -> bool:
        """Return True if the underlying IO object is available."""
        raise NotImplementedError

    def _write_bytes(self, frame: bytes):
        """Write raw bytes to the transport."""
        raise NotImplementedError

    def _close_io(self):
        """Close the underlying port/socket."""
        raise NotImplementedError

    def _read_loop(self, gen: int):
        """Background thread body — read data and feed _rx_buffer."""
        raise NotImplementedError

    def connect_device(self, device_path: str):
        """Establish connection to the given device/endpoint."""
        raise NotImplementedError

    def disconnect(self):
        """Gracefully disconnect."""
        raise NotImplementedError
